#! /usr/bin/python2
#-*- coding: utf-8 -*-

from PySide import QtGui, QtCore

import os
import sys
import shutil
import logging
from datetime import datetime
from zipfile import ZipFile
import subprocess

from ..methods.decorators import Decorator
from ..methods.dialogs import Dialogs
from ..methods.parser import PinguinoParser

from pinguino.dev import PinguinoLib


# Python3 compatibility
if os.getenv("PINGUINO_PYTHON") is "3":
    #Python3
    from configparser import RawConfigParser
    from os import makedirs
else:
    #Python2
    from ConfigParser import RawConfigParser

    #----------------------------------------------------------------------
    def makedirs(name, mode=0o777, exist_ok=False):
        """makedirs(name [, mode=0o777][, exist_ok=False])

        Super-mkdir; create a leaf directory and all intermediate ones.  Works like
        mkdir, except that any intermediate path segment (not just the rightmost)
        will be created if it does not exist. If the target directory already
        exists, raise an OSError if exist_ok is False. Otherwise no exception is
        raised.  This is recursive.

        """
        head, tail = os.path.split(name)
        if not tail:
            head, tail = os.path.split(head)
        if head and tail and not os.path.exists(head):
            try:
                makedirs(head, mode, exist_ok)
            except FileExistsError:
                # Defeats race condition when another thread created the path
                pass
            cdir = os.curdir
            if isinstance(tail, bytes):
                cdir = bytes(os.curdir, 'ASCII')
            if tail == cdir:           # xxx/newdir/. exists if xxx/newdir exists
                return
        try:
            os.mkdir(name, mode)
        except OSError:
            # Cannot rely on checking for EEXIST, since the operating system
            # could give priority to other errors like EACCES or EROFS
            if not exist_ok or not os.path.isdir(name):
                raise



########################################################################
class Project(object):
    """"""

    #----------------------------------------------------------------------
    def __init__(self):
        """"""
        self.connect(self.main.actionAdd_existing_directory, QtCore.SIGNAL("triggered()"), self.select_existing_directory)
        self.connect(self.main.actionNew_project, QtCore.SIGNAL("triggered()"), self.new_project)
        self.connect(self.main.actionNew_library, QtCore.SIGNAL("triggered()"), self.__show_library_template__)
        # self.connect(self.main.actionImport_library_project, QtCore.SIGNAL("triggered()"), self.import_library)
        self.connect(self.main.pushButton_newproject, QtCore.SIGNAL("clicked()"), self.new_project)
        self.connect(self.main.pushButton_openproject, QtCore.SIGNAL("clicked()"), self.open_project)
        self.connect(self.main.pushButton_newlibrary, QtCore.SIGNAL("clicked()"), self.__show_library_template__)
        self.connect(self.main.actionClose_project, QtCore.SIGNAL("triggered()"), self.close_project)
        self.connect(self.main.actionSave_project, QtCore.SIGNAL("triggered()"), self.save_project)
        self.connect(self.main.actionSave_project_as, QtCore.SIGNAL("triggered()"), self.save_project_as)
        self.connect(self.main.actionAdd_current_file, QtCore.SIGNAL("triggered()"), self.add_current_file)
        self.connect(self.main.actionOpen_project, QtCore.SIGNAL("triggered()"), self.open_project)
        self.connect(self.main.actionAdd_existing_file, QtCore.SIGNAL("triggered()"), self.select_existing_file)
        self.connect(self.main.treeWidget_projects, QtCore.SIGNAL("itemExpanded(QTreeWidgetItem*)"), self.expand_tree)
        self.connect(self.main.treeWidget_projects, QtCore.SIGNAL("itemDoubleClicked(QTreeWidgetItem*,int)"), self.open_from_tree)

        self.connect(self.main.pushButton_compilelibrary, QtCore.SIGNAL("clicked()"), self.compile_library)
        self.connect(self.main.pushButton_packagelibrary, QtCore.SIGNAL("clicked()"), self.package_library)

        self.connect(self.main.treeWidget_projects, QtCore.SIGNAL("itemChanged(QTreeWidgetItem*,int)"), self.update_check_status)
        # self.connect(self.main.treeWidget_projects, QtCore.SIGNAL("itemClicked(QTreeWidgetItem*,int)"), self.update_check_status)


    #----------------------------------------------------------------------
    @Decorator.show_tab("Project")
    @Decorator.alert_tab("Project")
    def open_project(self):

        self.project_saved = True
        file_project = Dialogs.set_open_file(self, exts="*.ppde PINGUINO")
        if file_project is None: return

        if file_project.endswith("PINGUINO"):
            self.import_library(file_project)
            return

        elif file_project.endswith(".ppde"):
            self.open_project_from_path(file_project)
            self.update_project_status()


    #----------------------------------------------------------------------
    @Decorator.show_tab("Project")
    @Decorator.alert_tab("Project")
    def open_project_from_path(self, filename):
        """"""
        try:
            self.project_saved = True
            self.ConfigProject = RawConfigParser()
            self.ConfigProject.readfp(open(filename, "r"))
            self.ConfigProject.filename = filename

            os.environ["PINGUINO_PROJECT"] = self.ConfigProject.get("Main", "name")

            project_name = self.reload_project(open_=True)


            if self.is_library():
                logging.debug("Opening \"{}\" library.".format(project_name))
            else:
                logging.debug("Opening \"{}\" project.".format(project_name))


            self.update_recents_projects(self.ConfigProject.filename)


            self.ide_close_all()
            if self.is_library():
                for path in self.get_libs_file():
                    self.ide_open_file_from_path(filename=path)

        except:
            Dialogs.error_message(self, "File seems to be corrupted!")

        self.save_recents()



    #----------------------------------------------------------------------
    def reload_project(self, open_=False):
        """"""

        if (not self.is_project()) and (not open_): return

        self.main.treeWidget_projects.clear()

        project_name = self.ConfigProject.get("Main", "name")
        self.main.treeWidget_projects.setHeaderLabel(project_name)

        if not self.ConfigProject.has_section("Ignore"):
            to_ignore = []
        elif not self.ConfigProject.options("Ignore"):
            to_ignore = []
        else:
            to_ignore = [self.ConfigProject.get("Ignore", option) for option in self.ConfigProject.options("Ignore")]

        inherits_status = self.get_inherits_from_project()

        for option in self.ConfigProject.options("Dirs"):
            if not self.ConfigProject.get("Dirs", option) in to_ignore:
                self.add_existing_directory(self.ConfigProject.get("Dirs", option), to_ignore, check_duplicated=False, inherits_status=inherits_status, open_=True, library=self.is_library())

        for path, checked in self.get_files_from_project().items():
            if path in to_ignore: continue
            self.add_existing_file(path, check_duplicated=False, check=checked, open_=True)

        self.update_project_status(project_name)
        return project_name


    #----------------------------------------------------------------------
    def select_existing_directory(self):
        """"""
        dir_project = Dialogs.set_open_dir(self)
        if dir_project in self.get_dirs_from_project(): return False

        if dir_project == self.get_default_project_dir():
            dir_project_tmp = os.path.join(dir_project, self.get_project_name())

            i = 1
            while os.path.exists(dir_project_tmp):
                dir_project_tmp = os.path.join(dir_project, self.get_project_name() + "-{}".format(i))
                i += 1

            dir_project = dir_project_tmp
            os.mkdir(dir_project)

        self.add_existing_directory(dir_project, inherits_status=True)
        return dir_project



    #----------------------------------------------------------------------
    def add_existing_directory(self, dir_project, to_ignore=[], check_duplicated=True, inherits_status=dict(), open_=False, library=False):
        """"""

        self.take_from_ignore(dir_project)
        if check_duplicated:
            if dir_project in self.get_dirs_from_project(): return

        # flags = QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsEditable|QtCore.Qt.ItemIsDragEnabled|QtCore.Qt.ItemIsUserCheckable|QtCore.Qt.ItemIsEnabled
        # flags = QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsDragEnabled|QtCore.Qt.ItemIsUserCheckable|QtCore.Qt.ItemIsEnabled

        if library is True:
            flags = QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsDragEnabled|QtCore.Qt.ItemIsEnabled
        else:
            flags = QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsDragEnabled|QtCore.Qt.ItemIsUserCheckable|QtCore.Qt.ItemIsEnabled


        parent = self.add_new_tree(os.path.basename(dir_project), self.main.treeWidget_projects, dir_project, flags)
        inherits = self.generate_tree(dir_project, parent, levels=50, flags=flags, to_ignore=to_ignore, inherits_status=inherits_status, library=library)
        parent.setExpanded(True)

        if not open_:
            self.add_dir_to_config(dir_project)
            self.add_inherits_to_config(inherits)


    #----------------------------------------------------------------------
    def add_current_file(self):
        """"""
        editor = self.get_tab().currentWidget()
        self.add_existing_file(editor.path)


    #----------------------------------------------------------------------
    def select_existing_file(self):
        """"""
        new_file = Dialogs.set_open_file(self)
        if new_file in self.get_files_from_project(): return
        self.add_existing_file(new_file)


    #----------------------------------------------------------------------
    def add_existing_file(self, path, check_duplicated=True, check=True, open_=False):
        """"""
        self.take_from_ignore(path)
        if check_duplicated:
            if path in self.get_files_from_project(): return

        filename = os.path.split(path)[1]
        parent = self.main.treeWidget_projects
        # flags = QtCore.Qt.ItemIsDragEnabled|QtCore.Qt.ItemIsUserCheckable|QtCore.Qt.ItemIsEnabled
        flags = QtCore.Qt.ItemIsSelectable|QtCore.Qt.ItemIsDragEnabled|QtCore.Qt.ItemIsUserCheckable|QtCore.Qt.ItemIsEnabled

        self.add_new_file_item(filename, parent, path, flags, check=check)

        if not open_:
            self.add_file_to_config(path)



    #----------------------------------------------------------------------
    def set_project_name(self):
        """"""
        project_name = Dialogs.get_text(self, "Project name", default="untitled_project")

        if project_name is False:
            return

        self.main.treeWidget_projects.setHeaderLabel(project_name)

        if not self.ConfigProject.has_section("Main"): self.ConfigProject.add_section("Main")
        options = self.ConfigProject.options("Main")
        self.ConfigProject.set("Main", "name", project_name)
        self.ConfigProject.set("Main", "library", False)

        self.main.widget_library.setVisible(False)

        self.save_project(default=True)

        logging.debug("Created \"{}\" project.".format(project_name))

        #os.environ["PINGUINO_PROJECT"] = project_name
        return project_name
        # self.main.treeWidget_projects.setHeaderLabel(project_name)


    #----------------------------------------------------------------------
    def set_library_name(self, project_name):
        """"""
        self.main.treeWidget_projects.setHeaderLabel(project_name)

        if not self.ConfigProject.has_section("Main"): self.ConfigProject.add_section("Main")
        options = self.ConfigProject.options("Main")
        self.ConfigProject.set("Main", "name", project_name)
        self.ConfigProject.set("Main", "library", True)

        self.main.widget_library.setVisible(True)

        self.save_project(default=True)

        logging.debug("Created \"{}\" library.".format(project_name))

        return project_name


    #----------------------------------------------------------------------
    def is_project(self):
        """"""
        if "PINGUINO_PROJECT" in os.environ:
            return bool(os.environ["PINGUINO_PROJECT"])
        else:
            return False


    #----------------------------------------------------------------------
    def is_library(self):
        """"""
        if os.environ.get("PINGUINO_PROJECT", False):
            if self.ConfigProject.has_option("Main", "library"):
                r = self.ConfigProject.get("Main", "library")
                if type(r) == str:
                    return r == "True"
                elif type(r) == bool:
                    return r
            else:
                self.ConfigProject.set("Main", "library", False)
                return False
        else:
            return False


    #----------------------------------------------------------------------
    def get_libs_file(self):
        """"""
        libs = []
        if self.ConfigProject.has_option("Main", "lib"):
            libs.append(self.ConfigProject.get("Main", "lib"))
        if self.ConfigProject.has_option("Main", "lib32"):
            libs.append(self.ConfigProject.get("Main", "lib32"))
        return libs


    #----------------------------------------------------------------------
    def get_project_name(self):
        """"""
        return self.ConfigProject.get("Main", "name")


    #----------------------------------------------------------------------
    def get_project_files(self):
        """"""
        #files = [file_ for file_, checked in self.get_files_from_project().iteritems() if checked]
        #files.extend([file_ for file_, checked in self.get_inherits_from_project().iteritems() if checked])
        files = [file_ for file_, checked in self.get_files_from_project().items() if checked]
        files.extend([file_ for file_, checked in self.get_inherits_from_project().items() if checked])
        return files


    # #----------------------------------------------------------------------
    # def get_main_upload(self):
        # """"""
        # header = self.main.treeWidget_projects.header()
        # filename = self.get_tab().currentWidget().path
        # return getattr(header, "main_upload", filename)


    #----------------------------------------------------------------------
    @Decorator.show_tab("Project")
    @Decorator.alert_tab("Project")
    def new_project(self):
        """"""
        self.project_saved = False
        self.ConfigProject = RawConfigParser()
        project_name = self.set_project_name()

        if project_name:
            Dialogs.info_message(self, QtGui.QApplication.translate("Dialogs", "Create or add an existing directory for project."))
            dirname = self.select_existing_directory()
            if dirname:
                self.create_project_template(dirname)
            self.update_project_status(project_name)

            self.reload_project()
            self.reload_project()


    #----------------------------------------------------------------------
    def create_project_template(self, dirname):
        """"""

        main_filename = os.path.join(dirname, 'main.pde')
        main_code = """/*-----------------------------------------------------
    Author: --<>
    Project: {}
    Date: {}
    Description:

    -----------------------------------------------------*/


    void setup(){{

    }}

    void loop(){{

        my_function();

    }}
    """.format(self.get_project_name(), datetime.now().strftime("%Y-%m-%d %H:%M:%S"))



        func_filename = os.path.join(dirname, 'functions.pde')
        func_code = """/*-----------------------------------------------------
    Author: --<>
    Project: {}
    Date: {}
    Description:

    -----------------------------------------------------*/


    void my_function(){{

    }}


    """.format(self.get_project_name(), datetime.now().strftime("%Y-%m-%d %H:%M:%S"))

        main_file = open(main_filename, mode="w")
        main_file.write(main_code)
        main_file.close()

        main_file = open(func_filename, mode="w")
        main_file.write(func_code)
        main_file.close()

        self.ide_open_file_from_path(filename=func_filename)
        self.ide_open_file_from_path(filename=main_filename)




    #----------------------------------------------------------------------
    def save_project_as(self):
        """"""
        self.project_saved = False
        self.save_project(silent=False)



    #----------------------------------------------------------------------
    def save_project(self, silent=False, default=False):
        """"""

        if default:
            default_path = os.path.join(os.getenv("PINGUINO_USER_PATH"), "projects")

            if not os.path.exists(default_path):
                makedirs(default_path, exist_ok=True)
            self.ConfigProject.filename = os.path.join(default_path, self.get_project_name()) + ".ppde"
            self.ConfigProject.write(open(self.ConfigProject.filename, "w"))
            self.project_saved = True
            return

        if self.project_saved:
            self.ConfigProject.write(open(self.ConfigProject.filename, "w"))

        elif not silent:
            save = Dialogs.set_save_file(self, self.get_project_name()+".ppde")
            if save[0]:
                path, filename = save
                self.ConfigProject.filename = path
                self.ConfigProject.write(open(self.ConfigProject.filename, "w"))
                self.project_saved = True

        self.update_recents_projects(self.ConfigProject.filename)


    #----------------------------------------------------------------------
    def add_dir_to_config(self, dirpath):
        """"""
        if not self.ConfigProject.has_section("Dirs"): self.ConfigProject.add_section("Dirs")
        options = self.ConfigProject.options("Dirs")

        if dirpath in self.get_dirs_from_project(): return

        index = 0
        while "dir_%d"%index in options:
            index += 1

        self.ConfigProject.set("Dirs", "dir_%d"%index, dirpath)
        self.save_project(silent=True)


    #----------------------------------------------------------------------
    def add_inherits_to_config(self, inherits):
        """"""
        if not self.ConfigProject.has_section("Inherits"): self.ConfigProject.add_section("Inherits")
        options = self.ConfigProject.options("Inherits")

        index = 0
        while "inherit__%d"%index in options:
            index += 1

        for path in inherits:
            if path in self.get_inherits_from_project(): continue
            self.ConfigProject.set("Inherits", "inherit_%d"%index, path)
            self.ConfigProject.set("Inherits", "inherit_%d_checked"%index, "1")
            index += 1
        self.save_project(silent=True)


    #----------------------------------------------------------------------
    def add_file_to_config(self, filepath):
        """"""
        if not self.ConfigProject.has_section("Files"): self.ConfigProject.add_section("Files")
        options = self.ConfigProject.options("Files")

        if filepath in self.get_files_from_project(): return

        index = 0
        while "file_%d"%index in options:
            index += 1

        self.ConfigProject.set("Files", "file_%d"%index, filepath)
        self.ConfigProject.set("Files", "file_%d_checked"%index, "1")

        self.save_project(silent=True)


    #----------------------------------------------------------------------
    def hide_element_to_config(self, path):
        """"""
        if not self.ConfigProject.has_section("Ignore"): self.ConfigProject.add_section("Ignore")
        options = self.ConfigProject.options("Ignore")

        index = 0
        while "ignore_%d"%index in options:
            index += 1

        self.ConfigProject.set("Ignore", "ignore_%d"%index, path)


        logging.debug("Removing \"{}\" from project.".format(path))
        for path_inherit, option in self.get_inherits_option_from_project().items():
            if path_inherit.startswith(path):
                self.ConfigProject.remove_option("Inherits", option)
                self.ConfigProject.remove_option("Inherits", option+"_checked")
                logging.debug("Removing \"{}\" from project.".format(path_inherit))


        for path_files, option in self.get_files_option_from_project().items():
            if path_files.startswith(path):
                self.ConfigProject.remove_option("Files", option)
                self.ConfigProject.remove_option("Files", option+"_checked")
                logging.debug("Removing \"{}\" from project.".format(path_files))




        self.save_project(silent=True)


    #----------------------------------------------------------------------
    def update_project_status(self, project="reload"):
        """"""
        if project == "reload":
            pass
        elif project:
            os.environ["PINGUINO_PROJECT"] = project
        else:
            os.environ["PINGUINO_PROJECT"] = ""

        self.main.treeWidget_projects.setVisible(bool(project))
        self.main.widget_noproject.setVisible(not bool(project))
        self.main.widget_library.setVisible(self.is_library())

        self.main.treeWidget_projects.contextMenuEvent = self.project_context_menu
        Decorator.update_toolbar()

        open_project = bool(os.environ["PINGUINO_PROJECT"])
        self.main.actionNew_project.setEnabled(not open_project)
        self.main.actionOpen_project.setEnabled(not open_project)
        self.main.actionSave_project.setEnabled(open_project)
        self.main.actionSave_project_as.setEnabled(open_project)
        self.main.actionClose_project.setEnabled(open_project)
        self.main.actionAdd_existing_directory.setEnabled(open_project)
        self.main.actionAdd_current_file.setEnabled(open_project)
        self.main.actionAdd_existing_file.setEnabled(open_project)
        self.main.actionAdd_new_file.setEnabled(open_project)
        self.main.actionAdd_new_blocks_file.setEnabled(open_project)
        # self.main.actionSet_current_file_as_main_upload.setEnabled(open_project)



    #----------------------------------------------------------------------
    def get_current_item(self):
        """"""
        return self.main.treeWidget_projects.currentItem()




    #----------------------------------------------------------------------
    def project_context_menu(self, event):
        menu = QtGui.QMenu()

        if hasattr(self.get_current_item(), "path"):
            if os.path.isfile(self.get_current_item().path):
                # option for files
                # menu.addAction("Set as main upload file", self.set_as_main_upload)
                menu.addAction("Remove file", self.remove_file)
            else:
                # option for dirs
                menu.addAction("Remove directory", self.remove_directory)

            # option for files and dirs
            menu.addAction("Remove exclude from project", self.remove_from_tree)

        # general options
        menu.addAction("Rename...", self.rename)
        menu.addSeparator()
        menu.addAction("Add existing directory...", self.select_existing_directory)
        menu.addAction("Add existing file...", self.select_existing_file)
        menu.addSeparator()
        if hasattr(self.get_current_item(), "path"):
            menu.addAction("Add new file...", self.add_new_file)
            menu.addAction("Add new blocks file...", self.add_new_blocks_file)
            menu.addAction("Add new directory...", self.add_new_directory)
        menu.addAction("Close project", self.close_project)
        menu.addAction("Open project...", self.open_project)
        menu.addSeparator()
        menu.addAction("Reload project", self.reload_project)

        menu.setStyleSheet("""
        QMenu {
            font-family: inherit;
            font-weight: normal;
            }

        """)

        menu.exec_(event.globalPos())


    # #----------------------------------------------------------------------
    # def set_as_main_upload(self):
        # """"""
        # item = self.get_current_item()
        # item.setForeground(0, QtGui.QColor("#FF0000"))

        # item.setText(0, "{} [main upload file]".format(item.text(0)))

        # if not self.ConfigProject.has_section("Main"): self.ConfigProject.add_section("Main")
        # self.ConfigProject.set("Main", "main_upload", item.path)
        # self.save_project(True)

        # header = self.main.treeWidget_projects.header()
        # header.main_upload = item.path


    #----------------------------------------------------------------------
    def remove_file(self):
        """"""
        item = self.get_current_item()
        os.remove(item.path)
        item.removeChild(item)


    #----------------------------------------------------------------------
    def remove_directory(self):
        """"""
        item = self.get_current_item()
        shutil.rmtree(item.path)
        item.removeChild(item)


    #----------------------------------------------------------------------
    def remove_from_tree(self):
        """"""
        item = self.get_current_item()
        self.hide_element_to_config(item.path)
        item.removeChild(item)


    #----------------------------------------------------------------------
    def take_from_ignore(self, path):
        """"""
        if self.ConfigProject.has_section("Ignore"):
            options = self.ConfigProject.options("Ignore")
            for option in options:
                if self.ConfigProject.get("Ignore", option) == path:
                    self.ConfigProject.remove_option("Ignore", option)


    #----------------------------------------------------------------------
    def get_default_project_dir(self):
        """"""
        dirs = self.get_dirs_from_project()
        if dirs:  return dirs[0]

        files = self.get_files_from_project().keys()
        if files: return files[0]

        # return QtCore.QDir.home().path()
        return os.getenv("PINGUINO_DEFAULT_FILES")


    #----------------------------------------------------------------------
    def get_dirs_from_project(self):
        """"""
        if not self.ConfigProject.has_section("Dirs"): return []
        if not self.ConfigProject.options("Dirs"): return []
        return [self.ConfigProject.get("Dirs", option) for option in self.ConfigProject.options("Dirs")]


    #----------------------------------------------------------------------
    def get_ignore_from_project(self):
        """"""
        if not self.ConfigProject.has_section("Ignore"): return []
        if not self.ConfigProject.options("Ignore"): return []
        return [self.ConfigProject.get("Ignore", option) for option in self.ConfigProject.options("Ignore")]


    #----------------------------------------------------------------------
    def get_files_from_project(self):
        """"""
        if not self.ConfigProject.has_section("Files"): return {}
        if not self.ConfigProject.options("Files"): return {}
        return {self.ConfigProject.get("Files", option):bool(int(self.ConfigProject.get("Files", "{}_checked".format(option))))
                for option in filter(lambda f:not f.endswith("_checked"), self.ConfigProject.options("Files"))}


    #----------------------------------------------------------------------
    def get_inherits_from_project(self):
        """"""
        if not self.ConfigProject.has_section("Inherits"): return {}
        if not self.ConfigProject.options("Inherits"): return {}
        return {self.ConfigProject.get("Inherits", option):bool(int(self.ConfigProject.get("Inherits", "{}_checked".format(option))))
                for option in filter(lambda f:not f.endswith("_checked"), self.ConfigProject.options("Inherits"))}


    #----------------------------------------------------------------------
    def get_files_option_from_project(self):
        """"""
        if not self.ConfigProject.has_section("Files"): return {}
        if not self.ConfigProject.options("Files"): return {}
        return {self.ConfigProject.get("Files", option):option
                for option in filter(lambda f:not f.endswith("_checked"), self.ConfigProject.options("Files"))}


    #----------------------------------------------------------------------
    def get_inherits_option_from_project(self):
        """"""
        if not self.ConfigProject.has_section("Inherits"): return {}
        if not self.ConfigProject.options("Inherits"): return {}
        return {self.ConfigProject.get("Inherits", option):option
                for option in filter(lambda f:not f.endswith("_checked"), self.ConfigProject.options("Inherits"))}



    #----------------------------------------------------------------------
    def update_check_status(self, item, int_):
        """"""
        status = item.checkState(int_)
        if status: status = "1"
        else: status = "0"

        if hasattr(item, "path"):

            path = item.path

            if not os.path.isfile(path):
                return

            if not path.endswith(".pde"):
                return

            inherits = self.get_inherits_option_from_project()
            files = self.get_files_option_from_project()

            if path in inherits:
                self.ConfigProject.set("Inherits", inherits[path]+"_checked", status)

            elif path in files:
                    self.ConfigProject.set("Files", files[path]+"_checked", status)

            else:
                self.update_new_file(path, status)
                # files = self.get_files_option_from_project()
                # self.ConfigProject.set("Files", files[path]+"_checked", status)

            self.save_project(silent=True)


    #----------------------------------------------------------------------
    def update_new_file(self, path, status):
        """
        Add new file and define their status.
        """
        dirs = self.get_dirs_from_project()
        for dir_ in dirs:
            if path.startswith(dir_):
                # add as inherits
                inherits = self.get_inherits_option_from_project()

                x = 0
                while "inherit_{}".format(x) in inherits.values():
                    x += 1

                self.ConfigProject.set("Inherits", "inherit_{}".format(x), path)
                self.ConfigProject.set("Inherits", "inherit_{}_checked".format(x), status)
                return

        files = self.get_files_option_from_project()
        x = 0
        while "files_{}".format(x) in files.values():
            x += 1
            self.ConfigProject.set("Files", "files_{}".format(x), path)
            self.ConfigProject.set("Files", "files_{}_checked".format(x), status)


        # add as file



    #----------------------------------------------------------------------
    def close_project(self):
        """"""
        name = self.get_project_name()
        self.project_saved = False
        self.main.treeWidget_projects.clear()
        self.update_project_status("")
        logging.debug("Closing \"{}\" project.".format(name))


    #----------------------------------------------------------------------
    def add_new_file(self):
        """"""
        path = self.get_current_item().path
        if os.path.isfile(path):
            path = os.path.dirname(self.get_current_item().path)

        filename = Dialogs.get_text(self, "New file", default=self.get_untitled_name(ext=".pde")[:-1])
        new_file = os.path.join(path, filename)
        nf = open(new_file, "w")
        nf.close()

        self.ide_open_file_from_path(filename=new_file)

        if new_file in self.get_files_option_from_project():
            self.add_existing_file(new_file)

        if self.is_project():
            self.reload_project()

    #----------------------------------------------------------------------
    def add_new_blocks_file(self):
        """"""
        path = self.get_current_item().path
        if os.path.isfile(path):
            path = os.path.dirname(self.get_current_item().path)


        filename = Dialogs.get_text(self, "New blocks file", default=self.get_untitled_name(ext=".gpde")[:-1])
        new_file = os.path.join(path, filename)
        nf = open(new_file, "w")
        nf.close()

        if self.get_current_item().path in self.get_files_option_from_project():
            self.add_existing_file(new_file)

        if self.is_project():
            self.reload_project()


    #----------------------------------------------------------------------
    def add_new_directory(self):
        """"""
        path = self.get_current_item().path
        if os.path.isfile(path):
            path = os.path.dirname(self.get_current_item().path)

        dirname = Dialogs.get_text(self, "New directory", default="untitled-dir")
        new_dir = os.path.join(path, dirname)
        makedirs(new_dir, exist_ok=True)

        if self.get_current_item().path in self.get_files_option_from_project():
            self.add_existing_directory(new_dir)

        if self.is_project():
            self.reload_project()

    #----------------------------------------------------------------------
    def rename(self):
        """"""
        # path = self.get_current_item().path
        path = getattr(self.get_current_item(), "path", "")
        name = Dialogs.get_text(self, "New name", default=os.path.basename(path))
        if not name: return

        new_name = os.path.join(os.path.dirname(path), name)
        files = self.get_files_option_from_project()
        os.rename(path, new_name)

        if path in files:
            self.ConfigProject.set("Files", files[path], new_name)

        if self.is_project():
            self.reload_project()


    #----------------------------------------------------------------------
    def update_recents_projects(self, filename):

        if filename in self.recent_projects:
            self.recent_projects.remove(filename)
        self.recent_projects.insert(0, filename)
        self.recent_projects = self.recent_projects[:10]

        self.update_recents_menu_project()


    #----------------------------------------------------------------------
    def update_recents_menu_project(self):

        self.main.menuRecentProjects.clear()
        for file_ in self.recent_projects:
            action = QtGui.QAction(self)
            filename = os.path.split(file_)[1]

            len_ = 40
            if len(file_) > len_:
                file_path_1 = file_[:int(len_/2)]
                file_path_2 = file_[int(-len_/2):]
                file_path = file_path_1 + "..." + file_path_2
            else: file_path = file_

            if os.path.isfile(file_):
                action.setText(filename+" ("+file_path+")")
                self.connect(action, QtCore.SIGNAL("triggered()"), self.ide_menu_recent_event(file_))
                action.ActionEvent = self.ide_menu_recent_event

                self.main.menuRecentProjects.addAction(action)

        self.main.menuRecentProjects.addSeparator()
        self.main.menuRecentProjects.addAction(QtGui.QApplication.translate("Dialogs", "Clear recent files"), self.clear_recents_projects)


    #----------------------------------------------------------------------
    def clear_recents_projects(self):

        self.main.menuRecentProjects.clear()
        self.main.menuRecentProjects.addSeparator()
        self.main.menuRecentProjects.addAction(QtGui.QApplication.translate("Dialogs", "Clear recent files"), self.clear_recents_projects)
        self.recent_projects = []


    #----------------------------------------------------------------------
    @Decorator.show_tab("Project")
    @Decorator.alert_tab("Project")
    def new_library(self, **kwargs):
        """"""
        library_name = kwargs["libname"]
        library_dir = kwargs["libpath"]

        self.project_saved = False
        self.ConfigProject = RawConfigParser()
        self.set_library_name(library_name)

        if library_dir:  #FIXME no necesario
            self.create_library_template(**kwargs)
        else:
            self.close_project()

        self.add_existing_directory(os.path.join(library_dir, library_name), inherits_status=True)
        self.update_project_status(library_name)

        if self.is_project():
            self.reload_project()

    #----------------------------------------------------------------------
    def create_library_template(self, libpath, libname, author, description, _8bit, _32bit):
        """"""
        library_dir = os.path.join(libpath, libname)
        example_dir = os.path.join(library_dir, "examples")
        # tests_dir = os.path.join(library_dir, "tests")
        pinguino_file = os.path.join(library_dir, "PINGUINO")
        # lib = os.path.join(library_dir, "{}.lib".format(libname))
        # self.ConfigProject.set("Main", "lib", lib)

        makedirs(library_dir, exist_ok=True)
        makedirs(example_dir, exist_ok=True)
        # makedirs(tests_dir, exist_ok=True)

        self.add_existing_directory(library_dir, inherits_status=True)

        lib_template = """/*-----------------------------------------------------
Author: {}--<>
Date: {}
Description:

-----------------------------------------------------*/


void my_private_function(){{


}}

PUBLIC u8 my_function(){{
    // This function could be used as {}.my_function

}}
""".format(author, datetime.now().strftime("%Y-%m-%d %H:%M:%S"), libname)

        if _8bit:
            lib = os.path.join(library_dir, "{}.lib".format(libname))
            self.ConfigProject.set("Main", "lib", lib)
            lib_file = open(lib, mode="w")
            lib_file.write(lib_template)
            lib_file.close()

        if _32bit:
            lib32 = os.path.join(library_dir, "{}.lib32".format(libname))
            self.ConfigProject.set("Main", "lib32", lib32)
            lib_file = open(lib32, mode="w")
            lib_file.write(lib_template)
            lib_file.close()

        example_file = open(os.path.join(example_dir, "example.pde"), mode="w")
        example_file.close()

        example_file = open(os.path.join(example_dir, "example.gpde"), mode="w")
        example_file.close()

        # test_file = open(os.path.join(tests_dir, "test.pde"), mode="w")
        # test_file.close()

        archs = []
        if _8bit: archs.append("8bit")
        if _32bit: archs.append("32bit")
        parse_lib = RawConfigParser()
        parse_lib.add_section("PINGUINO")
        parse_lib.set("PINGUINO", "name", libname)
        parse_lib.set("PINGUINO", "author", author)
        parse_lib.set("PINGUINO", "author_email", "")
        parse_lib.set("PINGUINO", "arch", ", ".join(archs))
        parse_lib.set("PINGUINO", "repository", "")
        parse_lib.set("PINGUINO", "description", description)
        parse_lib.set("PINGUINO", "url", "")
        parse_lib.set("PINGUINO", "version", "0.1")
        parse_lib.write(open(pinguino_file, "w"))

        self.ide_open_file_from_path(filename=lib)



    #----------------------------------------------------------------------
    def compile_library(self):
        """"""
        self.ide_save_all()
        self.update_library_project()

        lib_name = self.get_project_name()
        lib = self.ConfigProject.get("Main", "lib")

        archs = self.get_library_archs()
        filenames = []

        for arch in archs:

            lib_dir = os.path.dirname(lib)
            if not os.path.isdir(os.path.join(lib_dir, arch)):
                makedirs(os.path.join(lib_dir, arch), exist_ok=True)
            if not os.path.isdir(os.path.join(lib_dir, "pdl")):
                makedirs(os.path.join(lib_dir, "pdl"), exist_ok=True)

            define = os.path.join(lib_dir, arch, "{}.h".format(lib_name))
            userc = os.path.join(lib_dir, arch, "{}.c".format(lib_name))
            pdl = os.path.join(lib_dir, "pdl", "{}.pdl{}".format(lib_name, "" if arch=="p8" else "32"))

            filenames.append(define)
            filenames.append(userc)
            filenames.append(pdl)

            if arch == "p8":
                arch_num = 8
            elif arch == "p32":
                arch_num = 32

            libinstructions = self.pinguinoAPI.get_regobject_libinstructions(arch_num)
            self.pinguinoAPI.preprocess([lib], define_output=define, userc_output=userc, ignore_spaces=self.is_project(), libinstructions=libinstructions)
            # self.pinguinoAPI.preprocess([lib], libinstructions=libinstructions)

            userc_file = open(userc, mode="r")
            userc_content = userc_file.read()
            userc_file.close()

            include = "#include <{}.h>".format(lib_name)

            parser = PinguinoParser([{"filename":userc, "content":userc_content}])

            # functions = self.get_functions([{"filename":userc, "content":userc_content}])
            functions = parser.get_functions()
            userc_content = userc_content.split("\n")

            pdl_content = []

            for function in functions:
                if function["return"].startswith("PUBLIC"):
                    userc_content[function["line"]-1] = userc_content[function["line"]-1].replace("PUBLIC ", "", 1)
                    userc_content[function["line"]-1] = userc_content[function["line"]-1].replace(function["name"], "{}_{}".format(lib_name, function["name"]), 1)
                    pdl_content.append("{} {}#include <{}.c>\n".format("{}.{}".format(lib_name, function["name"]), "{}_{}".format(lib_name, function["name"]), lib_name))
                # else:
                    # userc_content[function["line"]-1] = userc_content[function["line"]-1].replace(function["name"], "{}_{}".format(lib_name, function["name"]), 1)


            header_content = """//------------------------------------------------------------------
// Pinguino Library source code generated automatically.
//
// {} for {}bit
//
// Created: {}
// by: {}
//
//  WARNING! All changes made in this file will be lost!
//------------------------------------------------------------------\n
""".format(lib_name, arch_num, datetime.now().strftime("%Y-%m-%d %H:%M:%S"), os.getenv("PINGUINO_FULLNAME"))


            userc_file = open(userc, mode="w")
            userc_file.writelines([header_content] + [include] + ["\n".join(userc_content)])
            userc_file.close()

            pdl_file = open(pdl, mode="w")
            pdl_file.writelines([header_content] + pdl_content + ["\n"])
            pdl_file.close()

        if self.is_project():
            self.reload_project()

        self.add_library_to_env(lib_dir)
        Dialogs.info_message(self, "Library \"{}\" compiled.".format(lib_name))

        for filename in filenames:
            logging.debug("Generated: \"{}\"".format(filename))




    #----------------------------------------------------------------------
    def add_library_to_env(self, lib_dir):
        """"""
        os_name = os.getenv("PINGUINO_OS_NAME")
        if os_name == "windows":
            separator = ";"
        elif os_name == "linux":
            separator = ":"
        #FIXME_OSX
        # elif os_name == "macosx":
            # separator = ":"


        if os.path.exists(os.path.join(lib_dir, "p32")):
            os.environ["PINGUINO_P32"] = os.path.join(lib_dir, "p32")

        if os.path.exists(os.path.join(lib_dir, "p8")):
            os.environ["PINGUINO_P8"] = os.path.join(lib_dir, "p8")

        if os.path.exists(os.path.join(lib_dir, "pdl")):
            # os.environ["PINGUINO_PDL"] = os.listdir(os.path.join(lib_dir, "pdl"))
            os.environ["PINGUINO_PDL"] = separator.join([os.path.join(lib_dir, "pdl", filename) for filename in os.listdir(os.path.join(lib_dir, "pdl")) if filename.endswith(".pdl") or filename.endswith(".pdl32")])


    #----------------------------------------------------------------------
    def get_library_archs(self):
        """"""
        archs = []
        if self.ConfigProject.has_option("Main", "lib"):
            archs.append("p8")
        if self.ConfigProject.has_option("Main", "lib32"):
            archs.append("p32")

        return archs


    #----------------------------------------------------------------------
    def get_library_version(self):
        """"""
        parse_lib = RawConfigParser()
        parse_lib.readfp(open(os.path.join(self.get_library_path(), "PINGUINO"), "r"))
        return parse_lib.get("PINGUINO", "version")


    #----------------------------------------------------------------------
    def get_library_path(self):
        """"""
        return os.path.dirname(self.get_library_libs()[0])


    #----------------------------------------------------------------------
    def get_library_libs(self):
        """"""
        libs = []
        if self.ConfigProject.has_option("Main", "lib"):
            libs.append(self.ConfigProject.get("Main", "lib"))
        if self.ConfigProject.has_option("Main", "lib32"):
            libs.append(self.ConfigProject.get("Main", "lib32"))

        return libs


    #----------------------------------------------------------------------
    def set_library_libs(self, lib=None, lib32=None):
        """"""
        if lib:
            self.ConfigProject.set("Main", "lib", lib)

        if lib32:
            self.ConfigProject.set("Main", "lib32", lib32)


    #----------------------------------------------------------------------
    def package_library(self):
        """"""
        cwd = os.getcwd()
        os.chdir(self.get_library_path())


        try:
            manifest = open(os.path.join(self.get_library_path(), "MANIFEST.in"), "w+")
            manifest.write(PinguinoLib.get_manifest_template())
            manifest.close()

            setup = open(os.path.join(self.get_library_path(), "setup.py"), "w+")
            setup.write(PinguinoLib.get_setup_template())
            setup.close()

            dest = "-d{}".format(os.path.join(self.get_library_path(), "packages"))
            p = subprocess.Popen([sys.executable, "setup.py", "sdist", dest, "--formats=zip"],
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
            out, err = p.communicate()
            self.write_log(out.decode())
            # self.write_log(err.decode())

            # if err:
                # raise Exception(err.decode())

            os.remove(os.path.join(self.get_library_path(), "MANIFEST.in"))
            os.remove(os.path.join(self.get_library_path(), "setup.py"))

            info = list(filter(lambda f:f.endswith(".egg-info"), os.listdir(self.get_library_path())))
            if info:
                try:
                    os.remove(os.path.join(self.get_library_path(), info[0]))
                except:
                    shutil.rmtree(os.path.join(self.get_library_path(), info[0]))

            if self.is_project():
                self.reload_project()
            Dialogs.info_message(self, "Package created in '{}'".format(os.path.join(self.get_library_path(), "packages")))


        except Exception as err:
            logging.error(err)

        os.chdir(cwd)

    # #----------------------------------------------------------------------
    # def package_library(self):
        # """"""
        # library_dir = self.get_library_path()
        # library_versioned_name = "{}-{}".format(self.get_project_name(), self.get_library_version())
        # package_name = "{}.zip".format(library_versioned_name)
        # package = ZipFile(os.path.join(library_dir, package_name), "w")

        # files = [
            # "pdl",
            # "examples",
            # # "{}.lib".format(self.get_project_name()),
            # # "{}.lib32".format(self.get_project_name()),
            # "p8",
            # "p32",
            # "PINGUINO",
            # "README.md",
            # "README",
        # ]

        # for target in files:
            # if os.path.exists(os.path.join(library_dir, target)):
                # package.write(os.path.join(library_dir, target), arcname=os.path.join(library_versioned_name, target))
                # logging.debug("Adding {} to {}".format(target, package_name))

                # if os.path.isdir(os.path.join(library_dir, target)):

                    # for root, dirs, files in os.walk(os.path.join(library_dir, target)):
                        # for file in files:
                            # package.write(os.path.join(root, file), arcname=os.path.join(library_versioned_name, os.path.split(root)[1], file))
                            # logging.debug("Adding {} to {}".format(os.path.join(root, file), package_name))


        # package.close()


    #----------------------------------------------------------------------
    def import_library(self, filename):
        """"""
        logging.debug("Loading library.")
        parce_pinguino = RawConfigParser()
        parce_pinguino.readfp(open(filename, "r"))

        library_dir = os.path.dirname(filename)
        library_name = parce_pinguino.get("PINGUINO", "name")

        self.project_saved = False
        self.ConfigProject = RawConfigParser()
        self.set_library_name(library_name)


        if "{}.lib".format(library_name) in os.listdir(library_dir):
            self.ConfigProject.set("Main", "lib", os.path.join(library_dir, "{}.lib".format(library_name)))
            self.ide_open_file_from_path(filename=os.path.join(library_dir, "{}.lib".format(library_name)))

        if "{}.lib32".format(library_name) in os.listdir(library_dir):
            self.ConfigProject.set("Main", "lib32", os.path.join(library_dir, "{}.lib32".format(library_name)))
            self.ide_open_file_from_path(filename=os.path.join(library_dir, "{}.lib32".format(library_name)))

        self.add_existing_directory(library_dir, inherits_status=True, library=True)
        self.update_project_status(library_name)

        #self.update_recents_projects(filename)

        #self.ide_open_file_from_path(filename=filename)


    #----------------------------------------------------------------------
    def update_library_project(self):
        """"""
        libs = []
        path = self.get_library_path()
        for file_ in os.listdir(path):
            if file_.endswith(".lib"):
                self.set_library_libs(lib=os.path.join(path, file_))
                libs.append("8bit")
            elif file_.endswith(".lib32"):
                self.set_library_libs(lib32=os.path.join(path, file_))
                libs.append("32bit")

        parse_lib = RawConfigParser()
        parse_lib.readfp(open(os.path.join(self.get_library_path(), "PINGUINO"), "r"))
        parse_lib.set("PINGUINO", "arch", ", ".join(set(libs)))
        self.ConfigProject.set("Main", "name", parse_lib.get("PINGUINO", "name"))
        # self.save_project(silent=False, default=False)
        parse_lib.write(open(os.path.join(self.get_library_path(), "PINGUINO"), "w"))




    #----------------------------------------------------------------------
    def get_library_kwargs(self):
        """"""
        path = self.get_default_project_dir()
        k = {}

        if os.path.exists(os.path.join(path, "p8")):
            k["p8"] = [os.path.join(path, "p8")]

        if os.path.exists(os.path.join(path, "p32")):
            k["p32"] = [os.path.join(path, "p32")]
        pdl = []
        for f in os.listdir(os.path.join(path, "pdl")):
            if f.endswith(".pdl") or  f.endswith(".pdl32"):
                pdl.append(os.path.join(path, "pdl", f))

        k["pdl"] = pdl

        return k


#!/usr/bin/env python
#-*- coding: utf-8 -*-

from . import constructor as Block
from .constructor import Kit

Kit = Kit()

########################################################################
class Pinguino(object):

    pinMode = lambda n:Kit.addFunct("Pinguino", "pinMode", 2, fill=["spin-int", "choice-io"])
    delay = lambda n:Kit.addFunct("Pinguino", "delay", 1, fill=["spin-int"])
    delayMicroseconds = lambda n:Kit.addFunct("Pinguino", "delayMicroseconds", 1, fill=["spin-int"])
    toggle = lambda n:Kit.addFunct("Pinguino", "toggle", 1, fill=["spin-int"])
    millis = lambda n:Kit.addFunct("Pinguino", "millis", 0, t="output")

    digitalWrite = lambda n:Kit.addFunct("Pinguino", "digitalWrite", 2, fill=["spin-int", "choice-hl"])
    analogWrite = lambda n:Kit.addFunct("Pinguino", "analogWrite", 2, fill=["spin-int", "slider"])
    digitalRead = lambda n:Kit.addFunct("Pinguino", "digitalRead", 1, fill=["spin-int"])
    analogRead = lambda n:Kit.addFunct("Pinguino", "analogRead", 1, fill=["spin-int"])

    digitalRead_o = lambda n:Kit.addFunct("Pinguino", "digitalRead", 1, t="output", fill=["spin-int"])
    analogRead_o = lambda n:Kit.addFunct("Pinguino", "analogRead", 1, t="output", fill=["spin-int"])

    digitalRead_b = lambda n:Kit.addFunct("Pinguino", "digitalRead", 1, t="output-bool", fill=["spin-int"])
    analogRead_b = lambda n:Kit.addFunct("Pinguino", "analogRead", 1, t="output-bool", fill=["spin-int"])



    OnTimer0 = lambda n:Kit.addFunct("Pinguino", "OnTimer0", 3, t="linear", fill=["spin-int", "interrupt_choice", "spin-int"])


    #----------------------------------------------------------------------
    def setup(self):

        st = Block.Function("Pinguino")
        st.addLabel("setup")
        st.addSyntax("()")
        return st.getBlock()


    #----------------------------------------------------------------------
    def loop(self):

        st = Block.Function("Pinguino")
        st.addLabel("loop")
        st.addSyntax("()")
        return st.getBlock()


    #----------------------------------------------------------------------
    def define(self):

        st = Block.Linear("Pinguino")
        st.addLabel("#define")
        st.addSyntax(" ")
        st.addSpace()
        st.addSyntax(" ")
        st.addSpace()
        st.addSyntax(" //")  #fix the ; at ent line, ugly but works!
        st.fillWith(["edit-value", "edit-value"])
        return st.getBlock()


    #----------------------------------------------------------------------
    def global_vars(self):

        st = Block.Function("Pinguino")
        st.addLabel("global")
        st.addSyntax(" variables")
        return st.getBlock()


    #----------------------------------------------------------------------
    def types(self):

        st = Block.Output("Output")
        st.addSpinChoice("char*", ["char*", "char", "int", "float", "u8", "u16", "u32"])
        st.addSyntax(" ")
        st.addEdit("Pinguino")
        return st.getBlock()



#!/usr/bin/env python
#-*- coding: utf-8 -*-

import os

from PySide import QtGui, QtCore


########################################################################
class Dialogs(object):

    #----------------------------------------------------------------------
    @classmethod
    def set_save_file(cls, parent, filename, dirpath=None):

        if dirpath is None: dirpath = parent.get_default_dir()
        if filename.endswith("*"): filename = filename[:-1]
        ext = os.path.splitext(os.path.split(filename)[1])[1]
        save_filename = QtGui.QFileDialog.getSaveFileName(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Save"),
                os.path.join(dirpath, filename),
                QtGui.QApplication.translate("Dialogs", "Pinguino files (*{});;All Files (*)".format(ext)))
        if save_filename: return save_filename[0], os.path.split(save_filename[0])[1]
        else: return None, None


    #----------------------------------------------------------------------
    @classmethod
    def set_open_file(cls, parent, exts="*.pde *.gpde *.ppde", dirpath=None):

        if dirpath is None:dirpath = parent.get_default_dir()
        open_file = QtGui.QFileDialog.getOpenFileName(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Open"),
                dirpath,
                QtGui.QApplication.translate("Dialogs", "Pinguino Files ({});;All Files (*)".format(exts)))
        if open_file[0]:
            return open_file[0]
        else: return None

    #----------------------------------------------------------------------
    @classmethod
    def set_open_files(cls, parent, exts="*.pde *.gpde *.ppde", dirpath=None):

        if dirpath is None:dirpath = parent.get_default_dir()
        open_files = QtGui.QFileDialog.getOpenFileNames(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Open"),
                dirpath,
                QtGui.QApplication.translate("Dialogs", "Pinguino Files ({});;All Files (*)".format(exts)))
        if open_files[0]:
            return open_files[0]
        else: return None


    #----------------------------------------------------------------------
    @classmethod
    def set_open_hex(cls, parent, dirpath=None):

        if dirpath is None:dirpath = parent.get_default_dir()
        open_file = QtGui.QFileDialog.getOpenFileName(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Select"),
                dirpath,
                QtGui.QApplication.translate("Dialogs", "Hex Files (*.hex);;All Files (*)"))
        if open_file: return open_file[0]
        else: return None


    #----------------------------------------------------------------------
    @classmethod
    def set_no_saved_file(cls, parent, filename):

        options = QtGui.QMessageBox.Yes | QtGui.QMessageBox.Discard | QtGui.QMessageBox.Cancel
        reply = QtGui.QMessageBox.question(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Save"),
                QtGui.QApplication.translate("Dialogs", "This file has not been saved,\nWould you like to do?\n\n")+filename[:-1],
                options)

        if reply == QtGui.QMessageBox.Yes: return True
        elif reply == QtGui.QMessageBox.Discard: return False
        else: return None


    #----------------------------------------------------------------------
    @classmethod
    def set_open_dir(cls, parent, dirpath=None):

        if dirpath is None:dirpath = parent.get_default_dir()
        open_dir = QtGui.QFileDialog.getExistingDirectory(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Open directory"),
                dirpath)

        if open_dir: return open_dir
        else: return None


    #----------------------------------------------------------------------
    @classmethod
    def file_duplicated(cls, parent, filename):

        QtGui.QMessageBox.information(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - File already open"),
                QtGui.QApplication.translate("Dialogs", "This file is already open in other tab.\n{}".format(filename)))
        return True


    #----------------------------------------------------------------------
    @classmethod
    def error_message(cls, parent, message):

        QtGui.QMessageBox.warning(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Error"),
                message)
        return True


    #----------------------------------------------------------------------
    @classmethod
    def confirm_message(cls, parent, message):

        options = QtGui.QMessageBox.Yes | QtGui.QMessageBox.Cancel
        reply = QtGui.QMessageBox.question(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Confirmation"),
                message,
                options)

        if reply == QtGui.QMessageBox.Yes: return True
        elif reply == QtGui.QMessageBox.Discard: return False
        else: return None


    #----------------------------------------------------------------------
    @classmethod
    def info_message(cls, parent, message):

        QtGui.QMessageBox.information(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Information"),
                message)
        return True


    #----------------------------------------------------------------------
    @classmethod
    def warning_message(cls, parent, message):

        QtGui.QMessageBox.warning(parent,
                                  os.getenv("PINGUINO_NAME") + QtGui.QApplication.translate("Dialogs", " - Warning"),
                                  message)

        return True


    #----------------------------------------------------------------------
    @classmethod
    def save_before_compile(cls, parent):

        QtGui.QMessageBox.information(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Save file first."),
                QtGui.QApplication.translate("Dialogs", "You must save the file before compiling."))
        return True


    #----------------------------------------------------------------------
    @classmethod
    def add_files_to_project(cls, parent):

        QtGui.QMessageBox.information(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Save file first."),
                QtGui.QApplication.translate("Dialogs", "You must add files to project before compiling."))
        return True


    #----------------------------------------------------------------------
    @classmethod
    def overwrite_file(cls, parent, filename, exist):

        msg_box = QtGui.QMessageBox(parent)
        msg_box.setIcon(QtGui.QMessageBox.Warning)
        msg_box.setWindowTitle(os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Error"))
        if exist:
            msg_box.setText(QtGui.QApplication.translate("Dialogs", "This file has changed on disk.")+"\n"+filename)
        else:
            msg_box.setText(QtGui.QApplication.translate("Dialogs", "This file has deleted/moved on disk.")+"\n"+filename)

        stdout = QtGui.QPushButton()
        stdout.setText(QtGui.QApplication.translate("Dialogs", "Overwrite"))

        if exist:
            ok = QtGui.QPushButton()
            ok.setText(QtGui.QApplication.translate("Dialogs", "Reload"))

        msg_box.addButton(stdout, QtGui.QMessageBox.RejectRole)
        if exist: msg_box.addButton(ok, QtGui.QMessageBox.NoRole)

        if exist: msg_box.setDefaultButton(ok)
        else: msg_box.setDefaultButton(stdout)

        msg_box.setStyleSheet("""
        font-family: inherit;
        font-weight: normal;

        """)

        reply = msg_box.exec_()


    #----------------------------------------------------------------------
    @classmethod
    def compilation_done(cls, parent):

        msg_box = QtGui.QMessageBox(parent)
        msg_box.setIcon(QtGui.QMessageBox.Information)
        msg_box.setWindowTitle(os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Compiled"))
        msg_box.setText(QtGui.QApplication.translate("Dialogs", "Compilation done!"))

        upload = QtGui.QPushButton()
        upload.setText(QtGui.QApplication.translate("Dialogs", "Upload now!"))

        ok = QtGui.QPushButton()
        ok.setText(QtGui.QApplication.translate("Dialogs", "Ok"))

        msg_box.addButton(upload, QtGui.QMessageBox.AcceptRole)
        msg_box.addButton(ok, QtGui.QMessageBox.NoRole)

        msg_box.setDefaultButton(ok)

        msg_box.setStyleSheet("""
        font-family: inherit;
        font-weight: normal;

        """)

        reply = msg_box.exec_()

        if reply == 0: return True
        elif reply == 1: return False


    #----------------------------------------------------------------------
    @classmethod
    def upload_done(cls, parent):

        QtGui.QMessageBox.information(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Upload done"),
                QtGui.QApplication.translate("Dialogs", "File sucessfully uploaded to Pinguino."))
        return True


    #----------------------------------------------------------------------
    @classmethod
    def upload_fail(cls, parent, message):

        msg_box = QtGui.QMessageBox(parent)
        msg_box.setIcon(QtGui.QMessageBox.Information)
        msg_box.setWindowTitle(os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Upload fail"))
        msg_box.setText(message)

        upload = QtGui.QPushButton()
        upload.setText(QtGui.QApplication.translate("Dialogs", "Try again!"))

        ok = QtGui.QPushButton()
        ok.setText(QtGui.QApplication.translate("Dialogs", "Cancel"))

        msg_box.addButton(upload, QtGui.QMessageBox.AcceptRole)
        msg_box.addButton(ok, QtGui.QMessageBox.NoRole)

        msg_box.setDefaultButton(upload)

        msg_box.setStyleSheet("""
        font-family: inherit;
        font-weight: normal;

        """)

        reply = msg_box.exec_()

        if reply == 0: return True
        elif reply == 1: return False


    #----------------------------------------------------------------------
    @classmethod
    def set_save_image(cls, parent, filename, dirpath=None):

        if dirpath is None:dirpath = parent.get_default_dir()
        file_name = QtGui.QFileDialog.getSaveFileName(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Save image"),
                os.path.join(dirpath, filename),
                QtGui.QApplication.translate("Dialogs", "Png files (*.png);;All Files (*)"))

        if file_name: return file_name[0]
        else: return None


    #----------------------------------------------------------------------
    @classmethod
    def get_text(self, parent, name, default=""):

        text, ok = QtGui.QInputDialog.getText(parent,
                os.getenv("PINGUINO_NAME")+QtGui.QApplication.translate("Dialogs", " - Input text"),
                name+":", QtGui.QLineEdit.Normal,
                default)
        if ok and text != '': return text
        else: return False



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/yeison/Documentos/Desarrollo/Pinguino/GitHub/pinguino-ide/pinguino/qtgui/gide/bloques/widgets/control_slider.ui'
#
# Created: Wed Mar 16 13:19:36 2016
#      by: pyside-uic 0.2.15 running on PySide 1.2.4
#
# WARNING! All changes made in this file will be lost!

from PySide import QtCore, QtGui

class Ui_Frame(object):
    def setupUi(self, Frame):
        Frame.setObjectName("Frame")
        Frame.resize(237, 36)
        Frame.setWindowTitle("")
        self.gridLayout = QtGui.QGridLayout(Frame)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.lineEdit_2 = QtGui.QLineEdit(Frame)
        self.lineEdit_2.setMaximumSize(QtCore.QSize(46, 16777215))
        font = QtGui.QFont()
        font.setFamily("Ubuntu Mono")
        font.setPointSize(15)
        font.setWeight(75)
        font.setBold(True)
        self.lineEdit_2.setFont(font)
        self.lineEdit_2.setStyleSheet("color: rgb(255, 255, 255);\n"
"background-color: rgba(255, 255, 255, 0);")
        self.lineEdit_2.setText("0000")
        self.lineEdit_2.setFrame(False)
        self.lineEdit_2.setReadOnly(True)
        self.lineEdit_2.setObjectName("lineEdit_2")
        self.gridLayout.addWidget(self.lineEdit_2, 0, 1, 1, 1)
        self.horizontalSlider = QtGui.QSlider(Frame)
        self.horizontalSlider.setCursor(QtCore.Qt.PointingHandCursor)
        self.horizontalSlider.setFocusPolicy(QtCore.Qt.NoFocus)
        self.horizontalSlider.setMaximum(1023)
        self.horizontalSlider.setOrientation(QtCore.Qt.Horizontal)
        self.horizontalSlider.setInvertedAppearance(False)
        self.horizontalSlider.setTickPosition(QtGui.QSlider.NoTicks)
        self.horizontalSlider.setTickInterval(128)
        self.horizontalSlider.setObjectName("horizontalSlider")
        self.gridLayout.addWidget(self.horizontalSlider, 0, 2, 1, 1)

        self.retranslateUi(Frame)
        QtCore.QMetaObject.connectSlotsByName(Frame)

    def retranslateUi(self, Frame):
        pass




#!/usr/bin/env python
#-*- coding: utf-8 -*-

import os
import shutil
import argparse
import re
import codecs

from .pinguino_tools import PinguinoTools
from .boards import boardlist as AllBoards

# Python3 compatibility
if os.getenv("PINGUINO_PYTHON") is "3":
    #Python3
    from os import makedirs
else:
    #Python2
    #----------------------------------------------------------------------
    def makedirs(name, mode=0o777, exist_ok=False):
        """makedirs(name [, mode=0o777][, exist_ok=False])

        Super-mkdir; create a leaf directory and all intermediate ones.  Works like
        mkdir, except that any intermediate path segment (not just the rightmost)
        will be created if it does not exist. If the target directory already
        exists, raise an OSError if exist_ok is False. Otherwise no exception is
        raised.  This is recursive.

        """
        head, tail = os.path.split(name)
        if not tail:
            head, tail = os.path.split(head)
        if head and tail and not os.path.exists(head):
            try:
                makedirs(head, mode, exist_ok)
            except FileExistsError:
                # Defeats race condition when another thread created the path
                pass
            cdir = os.curdir
            if isinstance(tail, bytes):
                cdir = bytes(os.curdir, 'ASCII')
            if tail == cdir:           # xxx/newdir/. exists if xxx/newdir exists
                return
        try:
            os.mkdir(name, mode)
        except OSError:
            # Cannot rely on checking for EEXIST, since the operating system
            # could give priority to other errors like EACCES or EROFS
            if not exist_ok or not os.path.isdir(name):
                raise



########################################################################
class Pinguino(PinguinoTools):


    #----------------------------------------------------------------------
    def __init__(self):
        """Pinguino

        Class for handle Pinguino from Python.
        """

        super(Pinguino, self).__init__()


    #----------------------------------------------------------------------
    def compile_file(self, file_path):
        """Compile file from path.

        Parameters
        ----------
        file_path: str, list
            Full absolute or relative path of file, list of files is project.

        See Also
        --------
        compile_string: Compile from string.
        """

        if type(file_path) == str:
            file_path = [file_path]

        data = self.verify(file_path)
        self.__compiled__ = data["verified"]
        self.__data__ = data


    #----------------------------------------------------------------------
    def compile_string(self, code, tmp_filename="temp_filename.pde", filenames=[]):
        """Compile from string.

        Parameters
        ----------
        code: str
            String with Pinguino Code for compile
        tmp_filename: str, optional
            Temporal filename used for compile the code.
        filenames: list of strings, optional
            Environment files for projects.

        See Also
        --------
        compile_file: Compile from path.

        Examples
        --------
        >>> code = "setup(){pinmode(0,OUTPUT);}loop(){toggle(0);delay(100);}"
        >>> compile_string(code)
        """

        tmp_dir = os.path.join(os.getenv("PINGUINO_USER_PATH"), "temp")
        if not os.path.exists(tmp_dir):
            makedirs(tmp_dir, exist_ok=True)
        tmp_file = os.path.join(tmp_dir, tmp_filename)
        file_ = codecs.open(tmp_file, "w", encoding="utf-8")
        file_.write(code)
        file_.close()
        self.compile_file([tmp_file] + filenames)


    #----------------------------------------------------------------------
    def upload(self):
        """Upload previous successfully compiled file.


        Returns
        -------
         uploaded: bool
            Is True if file was uploaded succesffully, otherwise is False.

        data: list
            List of messages reports.
        """

        data = self.__upload__()
        data = "\n".join(data)
        NO_UPLOADED_KEY_TEXT = ["Pinguino not found",
                                "Error",
                                "Upload not possible",
                                "No target found",
                                "Resource busy",
                                "Entity not found",
                                ]

        if False in [data.find(key) == -1 for key in NO_UPLOADED_KEY_TEXT]:
            if data.find("Error") != -1: data = data[data.find("Error"):]
            return False, data
        return True, data


    # #----------------------------------------------------------------------
    # def set_bootloader(self, bldr, memstart):
        # """Configure bootloader for current board selected.

        # Parameters
        # ----------
        # bldr: str
            # "noboot", "boot2" or "boot4"
        # memstart: int
            # 0, 0x2000 or 0x0C00

        # See Also
        # --------
        # set_board: Set the board to use.
        # """

        # if self.__current_board__.arch == 8:
            # self.__current_board__.bldr = bldr
            # self.__current_board__.memstart = memstart

    #----------------------------------------------------------------------
    def set_icsp(self):
        """"""
        if self.__current_board__.arch == 8:
            self.__current_board__.bldr = "noboot"
            self.__current_board__.memstart = 0


    #----------------------------------------------------------------------
    def set_8bit_compiler(self, compiler):
        """"""
        assert compiler.lower() in ["sdcc", "xc8"], "Compiler must be SDCC or XC8"
        self.SET_COMPILER_8BIT = compiler.lower()


    #----------------------------------------------------------------------
    def compiled(self):
        """Check if file was successfully compiled.

        Returns
        -------
        compiled: bool
            True if file was successfully compiled, otherwise is False.
        """

        return self.__compiled__


    #----------------------------------------------------------------------
    def get_errors(self):
        """Getter for errors while compiling.

        Returns
        -------
        errors: dict
            Dictionary with 3 types o errors (keys): compiling and linking


        See Also
        --------
        get_errors_linking: Get linking errors.
        get_errors_compiling_c: Get compiling C/C++ errors.
        get_errors_compiling_asm: Get compiling assembler errors.
        """

        errors = {}
        errors["compiling"] = self.__data__["compiling"]
        errors["linking"] = self.__data__["linking"]
        return errors


    #----------------------------------------------------------------------
    def get_errors_linking(self):
        """Getter for linking errors.

        Returns
        -------
        error: dict, None
            Dictionary with error message or None.

        See Also
        --------
        get_errors: Get errors while compiling.
        get_errors_compiling_c: Get compiling C/C++ errors.
        get_errors_compiling_asm: Get compiling assembler errors.
        """

        errors = self.__data__["linking"]
        if not errors: return None

        error = {}
        error["linking"] = errors
        error["line_numbers"] = []
        for err in errors:
            match = re.match(".*user.c:([\d]+):[\d]+: error.*", err)
            if not match is None: error["line_numbers"] += [int(match.groups()[0])]
        return error


    #----------------------------------------------------------------------
    def get_errors_compiling_c(self):
        """Getter for compiling C/C++ errors.

        Returns
        -------
        error: dict
            Dictionary with error message or None.

        See Also
        --------
        get_errors: Get errors while compiling.
        get_errors_linking: Get linking errors.
        get_errors_compiling_asm: Get compiling assembler errors.
        """

        errors = self.__data__["compiling"]["c"]
        if not errors: return None

        error = {}
        error["line_numbers"] = []
        error["complete_message"] = ""
        error["messages"] = {}
        for err in errors:
            error["line_numbers"] += [int(err["line_number"])]
            error["complete_message"] += err["line_number"] + ": " + err["message"]
            error["messages"][err["line_number"]] = err["message"]
        return error


    #----------------------------------------------------------------------
    def get_errors_compiling_asm(self):
        """Getter for compiling assembler errors.

        Returns
        -------
        error: str, None
            String with error message or None.

        See Also
        --------
        get_errors: Get errors while compiling.
        get_errors_linking: Get linking errors.
        get_errors_compiling_asm: Get compiling assembler errors.
        """

        errors = self.__data__["compiling"]["asm"]
        if not errors: return None

        if (type(errors) is dict) and ("error_symbols" in errors):
            return errors["error_symbols"]
        else:
            return errors


    #----------------------------------------------------------------------
    def get_result(self):
        """Getter log for post-compiling process.

        Returns
        -------
        result: dict
            Dictionary with info (keys) about: time, hex_file, code_size
        """

        assert self.__compiled__, "Must compile first."
        result = {}
        result["time"] = self.__data__["time"]
        result["hex_file"] = self.__data__["hex_file"]
        result["code_size"] = self.__data__["code_size"]
        return result


    #----------------------------------------------------------------------
    def get_hex(self):
        """Getter for hex file.

        Returns
        -------
        file: file
            Readonly hex file object.
        """

        assert self.__compiled__, "Must compile first."
        return open(self.__data__["hex_file"], "r")




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file '/home/yeison/Documentos/Desarrollo/Pinguino/GitHub/pinguino-ide/pinguino/qtgui/gide/bloques/frame_edit/frame_edit.ui'
#
# Created: Wed Mar 16 13:19:38 2016
#      by: pyside-uic 0.2.15 running on PySide 1.2.4
#
# WARNING! All changes made in this file will be lost!

from PySide import QtCore, QtGui

class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(380, 125)
        Form.setWindowTitle("")
        self.gridLayout = QtGui.QGridLayout(Form)
        self.gridLayout.setSpacing(0)
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.gridLayout.setObjectName("gridLayout")
        self.plainTextEdit = QtGui.QPlainTextEdit(Form)
        font = QtGui.QFont()
        font.setPointSize(12)
        self.plainTextEdit.setFont(font)
        self.plainTextEdit.setFrameShape(QtGui.QFrame.NoFrame)
        self.plainTextEdit.setPlainText("")
        self.plainTextEdit.setObjectName("plainTextEdit")
        self.gridLayout.addWidget(self.plainTextEdit, 1, 2, 1, 1)
        self.frame_8 = QtGui.QFrame(Form)
        self.frame_8.setMinimumSize(QtCore.QSize(0, 10))
        self.frame_8.setObjectName("frame_8")
        self.gridLayout.addWidget(self.frame_8, 0, 1, 1, 2)
        self.frame_7 = QtGui.QFrame(Form)
        self.frame_7.setMinimumSize(QtCore.QSize(10, 0))
        self.frame_7.setCursor(QtCore.Qt.OpenHandCursor)
        self.frame_7.setObjectName("frame_7")
        self.gridLayout.addWidget(self.frame_7, 1, 3, 1, 1)
        self.frame_6 = QtGui.QFrame(Form)
        self.frame_6.setMinimumSize(QtCore.QSize(0, 10))
        self.frame_6.setObjectName("frame_6")
        self.gridLayout.addWidget(self.frame_6, 2, 1, 1, 2)
        self.frame_4 = QtGui.QFrame(Form)
        self.frame_4.setObjectName("frame_4")
        self.gridLayout.addWidget(self.frame_4, 2, 3, 1, 1)
        self.frame_3 = QtGui.QFrame(Form)
        self.frame_3.setObjectName("frame_3")
        self.gridLayout.addWidget(self.frame_3, 2, 0, 1, 1)
        self.frame_2 = QtGui.QFrame(Form)
        self.frame_2.setObjectName("frame_2")
        self.gridLayout.addWidget(self.frame_2, 0, 3, 1, 1)
        self.frame_11 = QtGui.QFrame(Form)
        self.frame_11.setMinimumSize(QtCore.QSize(10, 0))
        self.frame_11.setObjectName("frame_11")
        self.gridLayout.addWidget(self.frame_11, 1, 1, 1, 1)
        self.frame_10 = QtGui.QFrame(Form)
        self.frame_10.setMinimumSize(QtCore.QSize(10, 0))
        self.frame_10.setObjectName("frame_10")
        self.gridLayout.addWidget(self.frame_10, 1, 0, 1, 1)
        self.frame = QtGui.QFrame(Form)
        self.frame.setObjectName("frame")
        self.gridLayout.addWidget(self.frame, 0, 0, 1, 1)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        pass




#!/usr/bin/env python
#-*- coding: utf-8 -*-

import os
import re
import logging

# Python3 compatibility
if os.getenv("PINGUINO_PYTHON") is "3":
    #Python3
    from configparser import RawConfigParser
else:
    #Python2
    from ConfigParser import RawConfigParser


########################################################################
class Config(RawConfigParser, object):

    #----------------------------------------------------------------------
    def __init__(self):
        """RawConfigParser based.

        A set of sections and options for storage configurations.

        [SECTION1]
        option11 = value11
        option12 = value12

        [SECTION2]
        option21 = (0, 1, 2)
        option22 = 3
        option23 = False

        [SECTION3]
        option3 = value3
        """

        super(Config, self).__init__()

        self.ide_config_file = os.path.join(os.getenv("PINGUINO_USER_PATH"), "pinguino.conf")
        self.verify_config_file()
        self.load_config()


    #----------------------------------------------------------------------
    def get_format_config(self, section, option):
        """Return formated configuration option.

        Parameters
        ----------
        section: str
            Section name.
        option: str
            Option name.

        Returns
        -------
        value: str, float, bool
            Return corret data type.
        """

        value = self.get(section, option)
        if type(value) != type(""): return value

        if re.match("([-\d]+)$", value):
            return int(value)

        if re.match("([.\d*e-]+)$", value):
            return float(value)

        elif re.match("(true)$", value.lower()):
            return True

        elif re.match("(false)$", value.lower()):
            return False

        else: return value


    #----------------------------------------------------------------------
    def config(self, section, option, default):
        """Read option or return default.

        Parameters
        ----------
        section: str
            Section name.
        option: str
            Option name.
        default: str, float, bool

        Returns
        -------
        option: str, float, bool
            If option exist then return them, otherwise return default value.
        """

        if self.has_section(section) and self.has_option(section, option):
            return self.get_format_config(section, option)
        else:
            self.set(section, option, default)
            return default


    #----------------------------------------------------------------------
    def set(self, section, option, value):
        """save option value.

        Parameters
        ----------
        section: str
            Section name.
        option: str
            Option name.
        value: str
            Value for option.
        """

        if not self.has_section(section): self.add_section(section)
        super(Config, self).set(section, option, value)


    #----------------------------------------------------------------------
    def verify_config_file(self):
        """Check if exist configuration file. otherwise create it.
        """

        if not os.path.isfile(self.ide_config_file):
            open(self.ide_config_file, "w").close()


    #----------------------------------------------------------------------
    def save_config(self):
        """Write configuration on configuration file.
        """

        self.write(open(self.ide_config_file, "w"))


    #----------------------------------------------------------------------
    def load_config(self):
        """Read configuration from configuration file.
        """

        self.readfp(open(self.ide_config_file, "r"))


    #----------------------------------------------------------------------
    def get_path(self, name, prefix=None):
        """Get corret formated path from config.

        Parameters
        ----------
        name: str
            Option name from configuration file.
        prefix: str, optional
            Path to add before the getted from configuration file, default is original path.

        Returns
        -------
        path: str
            Absolute path.
        """

        paths = self.config("Features", "pathstouse", "Paths")

        if self.has_option(paths, name):
            if prefix:
                return os.path.normpath(os.path.join(prefix, self.get(paths, name)[1:]))
            else:
                return os.path.normpath(self.get(paths, name))
        else:
            logging.warning("Missing '{}' for 'Paths' or 'PathsCustom' in config file: '{}'".format(name, self.ide_config_file))
            return ""


    #----------------------------------------------------------------------
    def get_recents(self, section="Recents"):
        """Getter for recents files from configuration.

        Parameters
        ----------
        section: str, optional
            Section for search recents.

        Returns
        -------
        files: list
            A list of paths for recents files.
        """

        if not self.has_section(section): self.add_section(section)
        options = self.options(section)

        files = []
        for option in options:
            if option.startswith("recent_") and self.get(section, option):
                files.append(self.get(section, option))

        return files


    #----------------------------------------------------------------------
    def get_recents_open(self, section="Recents"):
        """Getter for recents opened files from configuration.

        Parameters
        ----------
        section: str, optional
            Section for search recents opened.

        Returns
        -------
        files: list
            A list of paths for recents opened files.
        """

        if not self.has_section(section): self.add_section(section)
        options = self.options(section)

        files = []
        for option in options:
            if option.startswith("open_") and self.get(section, option):
                files.append(self.get(section, option))

        return files


    #----------------------------------------------------------------------
    def clear_recents_open(self, section="Recents"):
        """Remove recents opened files from configuration.

        Parameters
        ----------
        section: str, optional
            Section for remove recents opened.
        """

        if not self.has_section(section): self.add_section(section)
        options = self.options(section)

        for option in options:
            if option.startswith("open_"):
                self.remove_option(section, option)


    #----------------------------------------------------------------------
    def clear_recents(self, section="Recents"):
        """Remove recents files from configuration.

        Parameters
        ----------
        section: str, optional
            Section for remove recents.
        """

        if not self.has_section(section): self.add_section(section)
        options = self.options(section)

        for option in options:
            if option.startswith("recent_"):
                self.remove_option(section, option)



#!/usr/bin/env python
#-*- coding: utf-8 -*-
#-*- coding: iso-8859-15 -*-

"""---------------------------------------------------------------------
         _____ _____ _   _  _____ _    _ _____ _   _  ____             
        |  __ \_   _| \ | |/ ____| |  | |_   _| \ | |/ __ \            
        | |__) || | |  \| | |  __| |  | | | | |  \| | |  | |           
        |  ___/ | | | . ` | | |_ | |  | | | | | . ` | |  | |           
        | |    _| |_| |\  | |__| | |__| |_| |_| |\  | |__| |           
   ___  |_|   |_____|_| \_|\_____|\____/|_____|_| \_|\____/_           
  / _ \      | |   (_) |   | |  | |     | |               | |          
 | (_) |_____| |__  _| |_  | |  | |_ __ | | ___   __ _  __| | ___ _ __ 
  > _ <______| '_ \| | __| | |  | | '_ \| |/ _ \ / _` |/ _` |/ _ \ '__|
 | (_) |     | |_) | | |_  | |__| | |_) | | (_) | (_| | (_| |  __/ |   
  \___/      |_.__/|_|\__|  \____/| .__/|_|\___/ \__,_|\__,_|\___|_|   
                                  | |                                  
                                  |_|                                   
    Author:         Regis Blanchot <rblanchot@gmail.com>
    --------------------------------------------------------------------
    2013-11-13 - RB - first release   
    2015-09-08 - RB - fixed numBlocks > numBlocksMax when used with XC8
    2015-11-09 - RB - added PIC16F1459 ID
    2016-08-29 - RB - added usb.core functions
    2016-11-12 - RB - added python3 support (bytearray)
    2016-11-23 - RB - changed constant size DATABLOCKSIZE to variable writeBlockSize
    --------------------------------------------------------------------
    This library is free software you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
    MA  02110-1301  USA
---------------------------------------------------------------------"""

# This class is based on :
# - Diolan USB bootloader licenced (LGPL) by Diolan <http://www.diolan.com>
# - jallib USB bootloader licenced (BSD) by Albert Faber
# See also http://wiki.erazor-zone.de/wiki:projects:python:pyusb:pydoc
# Pinguino Device Descriptors : lsusb -v -d 04d8:feaa

import sys
import os
#import logging

from .uploader import baseUploader

class uploader8(baseUploader):
    """ upload .hex into pinguino device """

    # General Data Packet Structure (usbBuf)
    # ------------------------------------------------------------------
    #    __________________
    #    |    COMMAND     |   0       [CMD]
    #    |      LEN       |   1       [LEN]
    #    |     ADDRL      |   2       [        ]  [addrl]
    #    |     ADDRH      |   3       [ADR.pAdr]: [addrh]
    #    |     ADDRU      |   4       [        ]  [addru]
    #    |                |   5       [DATA]
    #    |                |
    #    .      DATA      .
    #    .                .
    #    |                |   62
    #    |________________|   63
    #
    # ------------------------------------------------------------------
    BOOT_CMD                        =    0
    BOOT_CMD_LEN                    =    1
    BOOT_ADDR_LO                    =    2
    BOOT_ADDR_HI                    =    3
    BOOT_ADDR_UP                    =    4
    BOOT_DATA_START                 =    5

    BOOT_REV1                       =    5
    BOOT_REV2                       =    6
    BOOT_DEV1                       =    7
    BOOT_DEV2                       =    8

    BOOT_VER_MINOR                  =    2
    BOOT_VER_MAJOR                  =    3

    BOOT_SIZE                       =    1

    # Bootloader commands
    # ------------------------------------------------------------------
    READ_VERSION_CMD                =    0x00
    READ_FLASH_CMD                  =    0x01
    WRITE_FLASH_CMD                 =    0x02
    ERASE_FLASH_CMD                 =    0x03
    #READ_EEDATA_CMD                =    0x04
    #WRITE_EEDATA_CMD               =    0x05
    #READ_CONFIG_CMD                =    0x06
    #WRITE_CONFIG_CMD               =    0x07
    RESET_CMD                       =    0xFF

    # USB Packet size
    # ------------------------------------------------------------------
    MAXPACKETSIZE                   =    64

    # bulk endpoints
    # ------------------------------------------------------------------
    IN_EP                           =    0x81    # endpoint for Bulk reads
    OUT_EP                          =    0x01    # endpoint for Bulk writes

    # configuration
    # ------------------------------------------------------------------

    # RB 2015-03-20 : moved to uploader.py
    #ACTIVE_CONFIG                   =    0x01
    #INTERFACE_ID                    =    0x00
    TIMEOUT                         =    10000       # 1200

    # Table with Microchip 8-bit USB devices
    # device_id:[PIC name]
    # ------------------------------------------------------------------

    devices_table = \
    {
        # 16F
        0x3020: ['16f1454'],
        0x3021: ['16f1455'],
        0x3023: ['16f1459'],
        0x3024: ['16lf1454'],
        0x3025: ['16lf1455'],
        0x3027: ['16lf1459'],

        # 18F
        0x4740: ['18f13k50'],
        0x4700: ['18lf13k50'],

        0x4760: ['18f14k50'],
        0x4720: ['18lf14k50'],

        0x2420: ['18f2450'],
        0x1260: ['18f2455'],
        0x2A60: ['18f2458'],
        0x4C00: ['18f24j50'],
        0x4CC0: ['18lf24j50'],
        0x5C60: ['18f24k50'],
        0x5CE0: ['18lf24k50'],

        0x1240: ['18f2550'],
        0x2A40: ['18f2553'],
        0x4C20: ['18f25j50'],
        0x4CE0: ['18lf25j50'],
        0x5C20: ['18f25k50'],
        0x5CA0: ['18lf25k50'],

        0x4C40: ['18f26j50'],
        0x4D00: ['18lf26j50'],

        0x5860: ['18f27j53'],

        0x1200: ['18f4450'],
        0x1220: ['18f4455'],
        0x2A20: ['18f4458'],
        0x4C60: ['18f44j50'],
        0x4D20: ['18lf44j50'],

        0x1200: ['18f4550'],
        0x2A00: ['18f4553'],
        0x4C80: ['18f45j50'],
        0x4D40: ['18lf45j50'],
        0x5C00: ['18f45k50'],
        0x5C80: ['18lf45k50'],

        0x4CA0: ['18f46j50'],
        0x4D60: ['18lf46j50'],

        0x58E0: ['18f47j53'],

        0x4100: ['18f65j50'],
        0x1560: ['18f66j50'],
        0x4160: ['18f66j55'],
        0x4180: ['18f67j50'],

        0x41A0: ['18f85j50'],
        0x41E0: ['18f86j50'],
        0x1F40: ['18f86j55'],
        0x4220: ['18f87j50']
    }

# ----------------------------------------------------------------------
    def usbWrite(self, device, usbBuf):
# ----------------------------------------------------------------------
        """ send bytes to the bootloader """

        if sys.version_info.major >= 3:
            usbBuf = str(bytearray(usbBuf))

        if self.PYUSB_USE_CORE:
            sent_bytes = device.write(self.OUT_EP, usbBuf, self.TIMEOUT)
        else:
            sent_bytes = device.bulkWrite(self.OUT_EP, usbBuf, self.TIMEOUT)

        return sent_bytes

# ----------------------------------------------------------------------
    def usbRead(self, device):
# ----------------------------------------------------------------------
        """ read bytes from the bootloader
            whatever is returned, USB packet size is always 64 bytes
            long (self.MAXPACKETSIZE) in high speed mode
        """

        if self.PYUSB_USE_CORE:
            usbBuf = device.read(self.IN_EP, self.MAXPACKETSIZE, self.TIMEOUT)
        else:
            usbBuf = device.bulkRead(self.IN_EP, self.MAXPACKETSIZE, self.TIMEOUT)

        if sys.version_info.major >= 3:
            return str(bytearray(usbBuf))
        else:
            return usbBuf
        
        #return usbBuf

# ----------------------------------------------------------------------
    def sendCommand(self, device, usbBuf):
# ----------------------------------------------------------------------
        """ send command to the bootloader """
        
        #self.add_report('[%s]' % ', '.join(map(hex, usbBuf)))
        sent_bytes = self.usbWrite(device, usbBuf)
        #self.add_report(str(sent_bytes))

        if sent_bytes != len(usbBuf):
            #self.add_report("Block not issued.")
            return self.ERR_USB_WRITE

        #self.add_report("Block issued without problem.")
        usbBuf = self.usbRead(device)
        #self.add_report('[%s]' % ', '.join(map(hex, usbBuf)))
        return usbBuf

# ----------------------------------------------------------------------
    def resetDevice(self, device):
# ----------------------------------------------------------------------
        """ reset device """
        
        #usbBuf = [self.RESET_DEVICE_CMD] * self.MAXPACKETSIZE
        usbBuf = [0] * self.MAXPACKETSIZE
        # command code
        usbBuf[self.BOOT_CMD] = self.RESET_CMD
        # write data packet
        #usbBuf = self.sendCommand(usbBuf)
        try:
            self.usbWrite(device, usbBuf)
        except:
            return self.ERR_USB_WRITE

        return self.ERR_NONE

# ----------------------------------------------------------------------
    def getVersion(self, device):
# ----------------------------------------------------------------------
        """ get bootloader version """
        
        usbBuf = [0] * self.MAXPACKETSIZE
        # command code
        usbBuf[self.BOOT_CMD] = self.READ_VERSION_CMD
        # write data packet and get response
        usbBuf = self.sendCommand(device, usbBuf)

        if usbBuf == self.ERR_USB_WRITE :
            return self.ERR_USB_WRITE

        # major.minor
        return str(usbBuf[self.BOOT_VER_MAJOR]) + "." + \
               str(usbBuf[self.BOOT_VER_MINOR])

# ----------------------------------------------------------------------
    def getDeviceID(self, device, board):
# ----------------------------------------------------------------------
        """ read 2-byte device ID from
            PIC18F : 0x3FFFFE
            PIC16F : 0x8005        """

        if (board.family == '16F'):
            # REVISION & DEVICE ID
            usbBuf = self.flashRead(device, 0x8005, 4)
            if usbBuf == self.ERR_USB_WRITE:
                return self.ERR_USB_WRITE, self.ERR_USB_WRITE
            rev1 = usbBuf[self.BOOT_REV1]
            rev2 = usbBuf[self.BOOT_REV2]
            device_rev = (int(rev2) << 8) + int(rev1)
            dev1 = usbBuf[self.BOOT_DEV1]
            dev2 = usbBuf[self.BOOT_DEV2]
            device_id  = (int(dev2) << 8) + int(dev1)
        else:
            # REVISION & DEVICE ID
            usbBuf = self.flashRead(device, 0x3FFFFE, 2)
            if usbBuf == self.ERR_USB_WRITE:
                return self.ERR_USB_WRITE, self.ERR_USB_WRITE
            #self.add_report("usbBuf = %s" % usbBuf)
            dev1 = usbBuf[self.BOOT_REV1]
            dev2 = usbBuf[self.BOOT_REV2]
            device_id  = (int(dev2) << 8) + int(dev1)
            device_id  = device_id & 0xFFE0
            device_rev = device_id & 0x001F

        return device_id, device_rev

# ----------------------------------------------------------------------
    def getDeviceName(self, device_id):
# ----------------------------------------------------------------------
        for n in self.devices_table:
            if n == device_id:
                return self.devices_table[n][0]
        return self.ERR_DEVICE_NOT_FOUND

# ----------------------------------------------------------------------
    def flashErase(self, device, address, numBlocks):
# ----------------------------------------------------------------------
        """ erase numBlocks of flash memory """
        
        usbBuf = [0] * self.MAXPACKETSIZE
        # command code
        usbBuf[self.BOOT_CMD] = self.ERASE_FLASH_CMD
        # number of blocks to erase
        usbBuf[self.BOOT_SIZE] = numBlocks
        #usbBuf[self.BOOT_SIZE] = int(numBlocks)
        # block address
        usbBuf[self.BOOT_ADDR_LO] = (address      ) & 0xFF
        usbBuf[self.BOOT_ADDR_HI] = (address >> 8 ) & 0xFF
        usbBuf[self.BOOT_ADDR_UP] = (address >> 16) & 0xFF
        # write data packet
        #return self.sendCommand(usbBuf)
        self.usbWrite(device, usbBuf)

# ----------------------------------------------------------------------
    def flashRead(self, device, address, length):
# ----------------------------------------------------------------------
        """ read a block of flash """
        
        usbBuf = [0] * self.MAXPACKETSIZE
        # command code
        usbBuf[self.BOOT_CMD] = self.READ_FLASH_CMD
        # size of block in bytes
        usbBuf[self.BOOT_CMD_LEN] = length
        # address of the block
        usbBuf[self.BOOT_ADDR_LO] = (address      ) & 0xFF
        usbBuf[self.BOOT_ADDR_HI] = (address >> 8 ) & 0xFF
        usbBuf[self.BOOT_ADDR_UP] = (address >> 16) & 0xFF
        # send request to the bootloader
        return self.sendCommand(device, usbBuf)
        #device.write(self.OUT_EP, usbBuf, self.TIMEOUT)
        #return device.read(self.IN_EP, self.BOOT_DATA_START + length, self.TIMEOUT)

# ----------------------------------------------------------------------
    def flashWrite(self, device, address, datablock):
# ----------------------------------------------------------------------
        """ write a block of code
            first 5 bytes are for block description (BOOT_CMD, BOOT_CMD_LEN and BOOT_ADDR)
            data block size should be of writeBlockSize bytes
            total length is then writeBlockSize + 5 """
            
        usbBuf = [0xFF] * self.MAXPACKETSIZE
        # command code
        usbBuf[self.BOOT_CMD] = self.WRITE_FLASH_CMD
        # size of block
        usbBuf[self.BOOT_CMD_LEN] = len(datablock)
        # block's address
        usbBuf[self.BOOT_ADDR_LO] = (address      ) & 0xFF
        usbBuf[self.BOOT_ADDR_HI] = (address >> 8 ) & 0xFF
        usbBuf[self.BOOT_ADDR_UP] = (address >> 16) & 0xFF
        # add data to the packet
        #self.txtWrite(hex(address))
        for i in range(len(datablock)):
            #self.txtWrite(hex(datablock[i]))
            usbBuf[self.BOOT_DATA_START + i] = datablock[i]
        # write data packet on usb device
        self.usbWrite(device, usbBuf)
        #print usbBuf
        #usbBuf = self.sendCommand(usbBuf)
        #print usbBuf
        #if usbBuf == self.ERR_USB_WRITE :
        #    self.txtWrite("Write error")

# ----------------------------------------------------------------------
    def writeHex(self, device, filename, board):
# ----------------------------------------------------------------------
        """ Parse the Hex File Format and send data to usb device """

        """
        [0]     Start code, one character, an ASCII colon ':'.
        [1:3]   Byte count, two hex digits, a number of bytes (hex digit pairs) in the data field. 16 (0x10) or 32 (0x20) bytes of data are the usual compromise values between line length and address overhead.
        [3:7]   Address, four hex digits, a 16-bit address of the beginning of the memory position for the data. Limited to 64 kilobytes, the limit is worked around by specifying higher bits via additional record types. This address is big endian.
        [7:9]   Record type, two hex digits, 00 to 05, defining the type of the data field.
        [9:*]   Data, a sequence of n bytes of the data themselves, represented by 2n hex digits.
        [*:*]   Checksum, two hex digits - the least significant byte of the two's complement of the sum of the values of all fields except fields 1 and 6 (Start code ":" byte and two hex digits of the Checksum). It is calculated by adding together the hex-encoded bytes (hex digit pairs), then leaving only the least significant byte of the result, and making a 2's complement (either by subtracting the byte from 0x100, or inverting it by XOR-ing with 0xFF and adding 0x01). If you are not working with 8-bit variables, you must suppress the overflow by AND-ing the result with 0xFF. The overflow may occur since both 0x100-0 and (0x00 XOR 0xFF)+1 equal 0x100. If the checksum is correctly calculated, adding all the bytes (the Byte count, both bytes in Address, the Record type, each Data byte and the Checksum) together will always result in a value wherein the least significant byte is zero (0x00).
                For example, on :0300300002337A1E
                03 + 00 + 30 + 00 + 02 + 33 + 7A = E2, 2's complement is 1E
        """

        # Addresses are doubled in the PIC16F HEX file
        if board.family == '16F':
            MEMSTART = board.memstart * 2
            MEMEND   = board.memend   * 2
        else:
            MEMSTART = board.memstart
            MEMEND   = board.memend

        #self.add_report("MEMSTART = 0x%X" % MEMSTART)
        #self.add_report("MEMEND   = 0x%X" % MEMEND)

        data        = []
        old_max_address = MEMSTART
        old_min_address = MEMEND
        max_address = 0
        min_address = 0
        address_Hi  = 0
        codesize    = 0

        # size of write block
        # ------------------------------------------------------------------

        if   "13k50" in board.proc :
            writeBlockSize = 8
        elif "14k50" in board.proc :
            writeBlockSize = 16
        else :
            writeBlockSize = 32

        # size of erase block
        # --------------------------------------------------------------

        # Pinguino 18Fx6j50 or 18Fx7j53
        # Erased block size is 1024-byte long
        if ("j" or "J") in board.proc :
            #self.add_report("x6j50, x6j53 or x7j53 chip")
            eraseBlockSize = 1024

        # Pinguino 16f145x, 18Fx455, 18Fx550 or 18Fx5k50
        # Erased block size is 64-byte long
        else:
            #self.add_report("x455, x550 or x5k50 chip")
            eraseBlockSize = 64

        # image of the whole PIC memory (above memstart)
        # --------------------------------------------------------------

        for i in range(MEMEND - MEMSTART):
            data.append(0xFF)

        # read hex file
        # --------------------------------------------------------------

        hexfile = open(filename,'r')
        lines = hexfile.readlines()
        hexfile.close()

        # calculate checksum and max_address
        # --------------------------------------------------------------

        for line in lines:
            byte_count = int(line[1:3], 16)
            # lower 16 bits (bits 0-15) of the data address
            address_Lo = int(line[3:7], 16)
            record_type= int(line[7:9], 16)

            # checksum calculation
            end = 9 + byte_count * 2 # position of checksum at end of line
            checksum = int(line[end:end+2], 16)
            cs = 0
            i = 1
            while i < end:
                cs = cs + (0x100 - int(line[i:i+2], 16) ) & 0xff # not(i)
                i = i + 2
            if checksum != cs:
                return self.ERR_HEX_CHECKSUM

            # extended linear address record
            if record_type == self.Extended_Linear_Address_Record:
                # upper 16 bits (bits 16-31) of the data address
                address_Hi = int(line[9:13], 16) << 16

            # data record
            elif record_type == self.Data_Record:

                address = address_Hi + address_Lo
                #self.add_report("address : 0x%X" % address)

                # min address
                if (address < old_min_address) and (address >= MEMSTART):
                    min_address = address
                    old_min_address = address
                    #self.add_report("min. address : 0x%X" % min_address)

                # max address
                if (address > old_max_address) and (address < MEMEND):
                    max_address = address + byte_count
                    old_max_address = address
                    #self.add_report("max. address : 0x%X" % max_address)

                # code size
                if (address >= MEMSTART) and (address < MEMEND):
                    codesize = codesize + byte_count
                    #self.add_report("0x%04X\t[%02d]\t[%05d]" % (address, byte_count, codesize))

                # data append
                for i in range(byte_count):
                    if ((address + i) < MEMEND):
                        #Caution : addresses are not always contiguous
                        #data.append(int(line[9 + (2 * i) : 11 + (2 * i)], 16))
                        #data[address - board.memstart + i] = int(line[9 + (2 * i) : 11 + (2 * i)], 16)
                        data[address - min_address + i] = int(line[9 + (2 * i) : 11 + (2 * i)], 16)
                        #self.add_report("0x%X=[%s]" % (address - min_address + i, int(line[9 + (2 * i) : 11 + (2 * i)], 16)))
                        #self.add_report("index=%d" % (address - min_address + i))

            # end of file record
            elif record_type == self.End_Of_File_Record:
                break

            # unsupported record type
            else:
                return self.ERR_HEX_RECORD

        # max_address must be divisible by writeBlockSize
        # --------------------------------------------------------------

        #min_address = min_address - eraseBlockSize - (min_address % eraseBlockSize)
        max_address = max_address + eraseBlockSize - (max_address % eraseBlockSize)
        if (max_address > MEMEND):
            max_address = MEMEND

        #self.add_report("min_address = 0x%X" % min_address)
        #self.add_report("max_address = 0x%X" % max_address)

        # erase memory from board.memstart to max_address
        # --------------------------------------------------------------

        numBlocksMax = (MEMEND      - MEMSTART) / eraseBlockSize
        numBlocks    = (max_address - MEMSTART) / eraseBlockSize
        #self.add_report("numBlocks    : %d" % numBlocks)
        #self.add_report("numBlocksMax : %d" % numBlocksMax)

        if  numBlocks > numBlocksMax:
            numBlocks = numBlocksMax
            #return self.ERR_USB_ERASE

        if  numBlocks < 256:
            self.flashErase(device, board.memstart, numBlocks)

        else:
            numBlocks = numBlocks - 255
            upperAddress = board.memstart + 255 * eraseBlockSize
            # from board.memstart to board.memstart + 255 x 64 = 0x3FC0
            self.flashErase(device, board.memstart, 255)
            # erase flash memory from board.memstart + 0x3FC0 to max_address
            self.flashErase(device, upperAddress, numBlocks)

        # write blocks of writeBlockSize bytes
        # --------------------------------------------------------------

        #for addr8 in range(board.memstart, max_address, writeBlockSize):
        for addr8 in range(min_address, max_address, writeBlockSize):
            #index = addr8 - board.memstart
            index = addr8 - min_address
            # the addresses are doubled in the PIC16F HEX file
            if board.family == '16F':
                addr16 = addr8 / 2
                self.flashWrite(device, addr16, data[index:index+writeBlockSize])
                #self.add_report("addr8=0x%X addr16=0x%X" % (addr8, addr16))
                #self.add_report("0x%X  [%s]" % (addr16, data[index:index+writeBlockSize]))
            else:
                self.flashWrite(device, addr8,  data[index:index+writeBlockSize])
                #self.add_report("0x%X  [%s]" % (addr8, data[index:index+writeBlockSize]))

        self.add_report("%d bytes written." % codesize)

        return self.ERR_NONE

# ----------------------------------------------------------------------
    #def writeHex(self):
    def uploadDevice(self, filename, board):
# ----------------------------------------------------------------------

        # check file to upload
        # --------------------------------------------------------------

        if filename == '':
            self.add_report("No program to write")
            self.closeDevice(device)
            return

        hexfile = open(filename, 'r')
        if hexfile == "":
            self.add_report("Unable to open %s" % filename)
            return
        hexfile.close()

        # search for a Pinguino board
        # --------------------------------------------------------------

        device = self.getDevice(board)

        if device == self.ERR_DEVICE_NOT_FOUND:
            self.add_report("Pinguino not found")
            self.add_report("If your device is connected,")
            self.add_report("press the Reset button to switch to bootloader mode.")
            return

        self.add_report("Pinguino found ...")

        device = self.initDevice(device)

        if device == self.ERR_USB_INIT1:
            self.add_report("Upload not possible")
            self.add_report("Try to restart the bootloader mode")
            return

        # find out the processor
        # --------------------------------------------------------------

        device_id, device_rev = self.getDeviceID(device, board)
        proc = self.getDeviceName(device_id)

        if proc == self.ERR_DEVICE_NOT_FOUND:
            self.add_report("Aborting: unknown PIC (id=0x%X)" % device_id)
            self.closeDevice(device)
            return
        else:
            self.add_report(" - with PIC%s (id=0x%X, rev=%x)" % (proc, device_id, device_rev))

        if proc != board.proc:
            self.add_report("Aborting: program compiled for %s but device has %s" % (board.proc, proc))
            self.closeDevice(device)
            return

        # find out flash memory size
        # --------------------------------------------------------------

        memfree = board.memend - board.memstart;
        self.add_report(" - with %d bytes free (%.2f/%d KB)" % (memfree, memfree/1024, board.memend/1024))
        self.add_report("   from 0x%05X to 0x%05X" % (board.memstart, board.memend))

        # find out bootloader version
        # --------------------------------------------------------------

        #product = device.getString(device.iProduct, 30)
        #manufacturer = device.getString(device.iManufacturer, 30)
        self.add_report(" - with USB bootloader v%s" % self.getVersion(device))

        # start writing
        # --------------------------------------------------------------

        self.add_report("Uploading user program ...")
        status = self.writeHex(device, filename, board)

        if status == self.ERR_HEX_RECORD:
            self.add_report("Aborting: record error")
            self.closeDevice(device)
            return

        elif status == self.ERR_HEX_CHECKSUM:
            self.add_report("Aborting: checksum error")
            self.closeDevice(device)
            return

        elif status == self.ERR_USB_ERASE:
            self.add_report("Aborting: erase error")
            self.closeDevice(device)
            return

        elif status == self.ERR_NONE:
            self.add_report(os.path.basename(filename) + " successfully uploaded")

        # reset and start start user's app.
        # --------------------------------------------------------------

            #self.txtWrite("Resetting ...")
            self.add_report("Starting user program ...")
            self.resetDevice(device)
            # Device can't be closed because it just has been reseted
            #self.closeDevice(device)
            return

        else:
            self.add_report("Aborting: unknown error")
            return
# ----------------------------------------------------------------------



#! /usr/bin/python2
#-*- coding: iso-8859-15 -*-

"""-------------------------------------------------------------------------
	Pinguino Universal Uploader

	Author:			Regis Blanchot <rblanchot@gmail.com>
	Last release:	2012-01-24

	This library is free software you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This library is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
-------------------------------------------------------------------------"""

# This class is based on :
# - Diolan USB bootloader licenced (LGPL) by Diolan <http://www.diolan.com>
# - jallib USB bootloader licenced (BSD) by Albert Faber
# See also PyUSB Doc. http://wiki.erazor-zone.de/wiki:projects:python:pyusb:pydoc
# Device Descriptors : lsusb -v -d 04d8:feaa

import sys
import os
#import usb			# checked in check.py

class uploaderDLN:
	""" upload .hex into pinguino device """

	# Bootloader commands
	# --------------------------------------------------------------------------

	DLN_WRITE_FLASH_CMD				=	0x01
	DLN_ERASE_FLASH_CMD				=	0x02
	DLN_GET_FW_VER_CMD				=	0x03
	DLN_RESET_CMD  					=	0x04
	DLN_READ_DEVID_CMD				=	0x05
	DLN_UNKNOWN_CMD					=	0xFF

	# boot command offset in bytes
	BOOT_CMD						=	0
	BOOT_ECHO						=	1
	BOOT_ADDR_LO					=	2
	BOOT_ADDR_HI					=	3
	BOOT_ADDR_UP					=	4
	BOOT_SIZE						=	5
	BOOT_CODE						=	6
	BOOT_VER_MAJOR					=	2
	BOOT_VER_MINOR					=	3
	BOOT_VER_SUBMINOR				=	4

	DLN_BLOCKSIZE					=	32 #64 - DLN_BOOT_CMD_SIZE

	BOOT_CMD_SIZE					=	64
	BOOT_DATA_SIZE					=	BOOT_CMD_SIZE - BOOT_CODE

	# HID endpoints
	DLN_IN_EP						=	0x81	# endpoint for HID reads
	DLN_OUT_EP						=	0x01	# endpoint for HID writes

	# configuration
	DLN_ACTIVE_CONFIG				=	0x01
	DLN_INTERFACE_ID				=	0x00
	DLN_TIMEOUT						=	500

	MAX_HID_RETRY					=	4

	# Table with supported USB devices (from JAL USB HID Bootloader)
	# device_id:[PIC name, flash size(in bytes), eeprom size (in bytes)]
	# --------------------------------------------------------------------------

	devices_table = \
		{
			0x4740: ['18f13k50'	, 0x02000, 0x80 ],
			0x4700: ['18lf13k50', 0x02000, 0x80 ],

			0x4760: ['18f14k50'	, 0x04000, 0xFF ],
			0x4720: ['18f14k50'	, 0x04000, 0xFF ],

			0x2420: ['18f2450'	, 0x04000, 0x00 ],
			0x1260: ['18f2455'	, 0x06000, 0xFF ],
			0x2A60: ['18f2458'	, 0x06000, 0xFF ],

			0x4C00: ['18f24J50'	, 0x04000, 0x00 ],
			0x4CC0: ['18lf24J50', 0x04000, 0x00 ],

			0x1240: ['18f2550'	, 0x08000, 0xFF ],
			0x2A40: ['18f2553'	, 0x08000, 0xFF ],

			0x4C20: ['18f25J50'	, 0x08000, 0x00 ],
			0x4CE0: ['18lf25J50', 0x08000, 0x00 ],

			0x4C40: ['18f26J50'	, 0x10000, 0x00 ],
			0x4D00: ['18lf26J50', 0x10000, 0x00 ],

			0x1200: ['18f4450'	, 0x04000, 0x00 ],  #FIXME: Duplicate key 0x1200
			0x1220: ['18f4455'	, 0x06000, 0x00 ],
			0x2A20: ['18f4458'	, 0x06000, 0xFF ],

			0x4C60: ['18f44J50'	, 0x04000, 0x00 ],
			0x4D20: ['18lf44J50', 0x04000, 0x00 ],

			0x1200: ['18f4550'	, 0x08000, 0xFF ],  #FIXME: Duplicate key 0x1200
			0x2A00: ['18f4553'	, 0x08000, 0xFF ],

			0x4C80: ['18f45J50'	, 0x08000, 0x00 ],
			0x4D40: ['18lf45J50', 0x08000, 0x00 ],

			0x4CA0: ['18f46J50'	, 0x10000, 0x00 ],
			0x4D60: ['18f46J50'	, 0x10000, 0x00 ],

			0x4100: ['18f65J50'	, 0x08000, 0x00 ],
			0x1560: ['18f66J50'	, 0x10000, 0x00 ],
			0x4160: ['18f66J55'	, 0x18000, 0x00 ],
			0x4180: ['18f67J50'	, 0x20000, 0x00 ],
			0x41A0: ['18f85J50'	, 0x08000, 0x00 ],
			0x41E0: ['18f86J50'	, 0x10000, 0x00 ],
			0x1F40: ['18f86J55'	, 0x18000, 0x00 ],
			0x4220: ['18f87J50'	, 0x20000, 0x00 ]
		}

# ------------------------------------------------------------------------------
	def initDevice(self, device):
# ------------------------------------------------------------------------------
		""" init pinguino device """
		conf = device.configurations[0]
		iface = conf.interfaces[0][0]
		handle = device.open()
		if handle:
			if sys.platform == 'win32':
				handle.setConfiguration(conf)
				handle.claimInterface(iface)
				handle.setAltInterface(iface)
			if sys.platform == 'linux2':
				handle.detachKernelDriver(iface.interfaceNumber)
				handle.setConfiguration(conf)
				handle.claimInterface(iface)
				handle.setAltInterface(iface)
			return handle
		return self.ERR_USB_INIT1
# ------------------------------------------------------------------------------
	def usbWrite(self, usbBuf):
# ------------------------------------------------------------------------------
		"""
		controlMsg(requestType, request, buffer, value=0, index=0, timeout=100) -> bytesWritten|buffer
		Performs a control request to the default control pipe on a device.
		Arguments:
			requestType: specifies the direction of data flow, the type of request
				and the recipient.
				= Host to Device, Class type, Interface recipient
				= usb.TYPE_CLASS | usb.RECIP_INTERFACE | usb.ENDPOINT_OUT
				= 0x21
			request: specifies the request.
				= SET_REPORT
				= 0x9
			buffer: if the transfer is a write transfer, buffer is a sequence
				with the transfer data, otherwise, buffer is the number of
				bytes to read.
			value: specific information to pass to the device. (default: 0)
				= 0
			index: specific information to pass to the device. (default: 0)
				= 0
			timeout: operation timeout in miliseconds. (default: 100)
		Returns the number of bytes written.
		"""
		sent_bytes = handle.controlMsg(0x21, 0x09, usbBuf, 0x00, 0x00, self.DLN_TIMEOUT)
		if sent_bytes == len(usbBuf):
			return self.ERR_NONE
		else:
			return self.ERR_USB_WRITE
# ------------------------------------------------------------------------------
	def transaction(self, handle, usbBuf):
# ------------------------------------------------------------------------------
		"""
		Write a data packet to currently-open USB device
		Return 64 bytes from the device

		interruptRead(endpoint, size, timeout=100) -> buffer
		Performs a interrupt read request to the endpoint specified.
		Arguments:
			endpoint: endpoint number.
			size: number of bytes to read.
			timeout: operation timeout in miliseconds. (default: 100)
		Returns a tuple with the data read.
		"""
		retry = 0
		while retry < self.MAX_HID_RETRY:
			status = self.usbWrite(handle, usbBuf)
			if status == self.ERR_NONE:
				return handle.interruptRead(1, 64, self.DLN_TIMEOUT)
				#return handle.interruptRead(1, len(usbBuf), self.DLN_TIMEOUT)
			else:
				retry = retry + 1
		return self.ERR_USB_WRITE
# ------------------------------------------------------------------------------
	def reset(self, handle):
# ------------------------------------------------------------------------------
		""" reset device """
		usbBuf = [0] * 64
		# command code
		usbBuf[self.BOOT_CMD] = self.DLN_RESET_CMD
		# write data packet
		return self.usbWrite(handle, usbBuf)
# ------------------------------------------------------------------------------
	def getVersion(self, handle):
# ------------------------------------------------------------------------------
		""" get bootloader version """
		usbBuf = [0] * 64
		# command code
		usbBuf[self.BOOT_CMD] = self.DLN_GET_FW_VER_CMD
		# write data packet and get response
		usbBuf = self.transaction(handle, usbBuf)
		if usbBuf == self.ERR_USB_WRITE:
			return self.ERR_USB_WRITE
		else:
			# major.minor.subminor
			return	str(usbBuf[self.BOOT_VER_MAJOR]) + "." + \
					str(usbBuf[self.BOOT_VER_MINOR]) + "." + \
					str(usbBuf[self.BOOT_VER_SUBMINOR])
# ------------------------------------------------------------------------------
	def getDeviceID(self, handle):
# ------------------------------------------------------------------------------
		""" read Device ID """
		usbBuf = [0] * 64
		# command code
		usbBuf[self.BOOT_CMD] = self.DLN_READ_DEVID_CMD
		# write data packet and get response
		usbBuf = self.transaction(handle, usbBuf)
		if usbBuf == self.ERR_USB_WRITE:
			return self.ERR_USB_WRITE
		else:
			dev1 = usbBuf[self.BOOT_CODE + 0]
			dev2 = usbBuf[self.BOOT_CODE + 1]
			device_id = ( int(dev2) << 8 ) + int(dev1)
			deivce_rev = device_id & 0x001F
			# mask revision number
			return device_id  & 0xFFE0
# ------------------------------------------------------------------------------
	def getDeviceFlash(self, device_id, board):
# ------------------------------------------------------------------------------
		for n in self.devices_table:
			if n == device_id:
				return self.devices_table[n][1] - board.memstart
		return self.ERR_DEVICE_NOT_FOUND
# ------------------------------------------------------------------------------
	def getDeviceName(self, device_id):
# ------------------------------------------------------------------------------
		for n in self.devices_table:
			if n == device_id:
				return self.devices_table[n][0]
		return self.ERR_DEVICE_NOT_FOUND
# ------------------------------------------------------------------------------
	def eraseFlash(self, handle, address, size64):
# ------------------------------------------------------------------------------
		""" erase n * 64-byte blocks of flash memory """
		usbBuf = [0] * 64
		# command code
		usbBuf[self.BOOT_CMD] = self.DLN_ERASE_FLASH_CMD
		# echo is used to link between command and response
		usbBuf[self.BOOT_ECHO] = 123
		# block address (must be divisible by 64)
		usbBuf[self.BOOT_ADDR_LO] = (address      ) & 0xFF
		usbBuf[self.BOOT_ADDR_HI] = (address >> 8 ) & 0xFF
		usbBuf[self.BOOT_ADDR_UP] = (address >> 16) & 0xFF
		# size in 64 byte blocks
		usbBuf[self.BOOT_SIZE] = size64
		# write data packet and get response
		#print usbBuf
		usbBuf = self.transaction(handle, usbBuf)
		#print usbBuf
# ------------------------------------------------------------------------------
	def writeFlash(self, handle, address, block):
# ------------------------------------------------------------------------------
		""" write a block of code """
		usbBuf = [0xFF] * 64
		usbBuf[self.BOOT_CMD] = self.DLN_WRITE_FLASH_CMD
		# echo is used to link between command and response
		usbBuf[self.BOOT_ECHO] = 123
		# block's address (must be divisible by 2)
		#print hex(address)
		usbBuf[self.BOOT_ADDR_LO] = (address      ) & 0xFF
		usbBuf[self.BOOT_ADDR_HI] = (address >> 8 ) & 0xFF
		usbBuf[self.BOOT_ADDR_UP] = (address >> 16) & 0xFF
		# size must be divisible by 8
		size = len(block)
		if (size % 8 != 0):
			size = (size / 8 + 1) * 8
		usbBuf[self.BOOT_SIZE] = size
		# add data to the packet
		for i in range(size):
			usbBuf[self.BOOT_CODE + i] = block[i]
		# write data packet on usb device
		#print usbBuf
		usbBuf = self.transaction(handle, usbBuf)
		#print usbBuf
# ------------------------------------------------------------------------------
	def hexWrite(self, handle, filename, board):
# ------------------------------------------------------------------------------
		""" Parse the Hex File Format and send data to usb device """

		"""
		[0]		Start code, one character, an ASCII colon ':'.
		[1:3]	Byte count, two hex digits, a number of bytes (hex digit pairs) in the data field. 16 (0x10) or 32 (0x20) bytes of data are the usual compromise values between line length and address overhead.
		[3:7]	Address, four hex digits, a 16-bit address of the beginning of the memory position for the data. Limited to 64 kilobytes, the limit is worked around by specifying higher bits via additional record types. This address is big endian.
		[7:9]	Record type, two hex digits, 00 to 05, defining the type of the data field.
		[9:*]	Data, a sequence of n bytes of the data themselves, represented by 2n hex digits.
		[*:*]	Checksum, two hex digits - the least significant byte of the two's complement of the sum of the values of all fields except fields 1 and 6 (Start code ":" byte and two hex digits of the Checksum). It is calculated by adding together the hex-encoded bytes (hex digit pairs), then leaving only the least significant byte of the result, and making a 2's complement (either by subtracting the byte from 0x100, or inverting it by XOR-ing with 0xFF and adding 0x01). If you are not working with 8-bit variables, you must suppress the overflow by AND-ing the result with 0xFF. The overflow may occur since both 0x100-0 and (0x00 XOR 0xFF)+1 equal 0x100. If the checksum is correctly calculated, adding all the bytes (the Byte count, both bytes in Address, the Record type, each Data byte and the Checksum) together will always result in a value wherein the least significant byte is zero (0x00).
				For example, on :0300300002337A1E
				03 + 00 + 30 + 00 + 02 + 33 + 7A = E2, 2's complement is 1E
		"""

		data		= []
		old_address = 0
		max_address	= 0
		address_Hi	= 0
		codesize	= 0

		# read hex file
		# ----------------------------------------------------------------------

		fichier = open(filename,'r')
		lines = fichier.readlines()
		fichier.close()

		# 1st pass : calculate checksum and max_address
		# ----------------------------------------------------------------------

		for line in lines:
			byte_count = int(line[1:3], 16)
			address_Lo = int(line[3:7], 16) # lower 16 bits (bits 0-15) of the data address
			record_type= int(line[7:9], 16)

			address = (address_Hi << 16) + address_Lo

			# checksum calculation
			end = 9 + byte_count * 2 # position of checksum at end of line
			checksum = int(line[end:end+2], 16)
			cs = 0
			i = 1
			while i < end:
				cs = cs + (0x100 - int(line[i:i+2], 16) ) & 0xff # not(i)
				i = i + 2
			if checksum != cs:
				return self.ERR_HEX_CHECKSUM

			# extended linear address record
			if record_type == self.Extended_Linear_Address_Record:
				address_Hi = int(line[9:13], 16) << 16 # upper 16 bits (bits 16-31) of the data address
				address = (address_Hi << 16) + address_Lo

			# code size
			if address >= board.memstart:
				codesize = codesize + byte_count

			# max address
			if (address > old_address) and (address < board.memend):
				max_address = address + byte_count
				old_address = address

		max_address = max_address + 64 - (max_address % 64)
		#print memstart, max_address, memend

		# fill data sequence with 0xFF
		# ----------------------------------------------------------------------

		for i in range(max_address - board.memstart):
			data.append(0xFF)

		# 2nd pass : parse bytes from line into data
		# ----------------------------------------------------------------------

		address_Hi	= 0

		for line in lines:
			byte_count = int(line[1:3], 16)
			address_Lo = int(line[3:7], 16) # four hex digits
			record_type= int(line[7:9], 16)

			# 32-bit address
			address = (address_Hi << 16) + address_Lo
			#print hex(address)

			# data record
			if record_type == self.Data_Record:
				if (address >= board.memstart) and (address < board.memend):
					#print hex(address)
					for i in range(byte_count):
						#print address - memstart + i
						#print int(line[9 + (2 * i) : 11 + (2 * i)], 16)
						data[address - board.memstart + i] = int(line[9 + (2 * i) : 11 + (2 * i)], 16)

			# end of file record
			elif record_type == self.End_Of_File_Record:
				break

			# extended linear address record
			elif record_type == self.Extended_Linear_Address_Record:
				address_Hi = int(line[9:13], 16) # upper 16 bits (bits 16-31) of the data address
				#print "address_Hi = " + hex(address_Hi)

			# unsupported record type
			else:
				return self.ERR_HEX_RECORD

		# erase memory from memstart to max_address
		# ----------------------------------------------------------------------

		size64 = (max_address - board.memstart) / 64
		if size64 > 511:
			return ERR_USB_ERASE
		if size64 < 256:
			self.eraseFlash(handle, board.memstart, size64)
		else:
			# erase flash memory from memstart to memstart + 0x4000
			self.eraseFlash(handle, board.memstart, 255)
			# erase flash memory from memstart + 0x4000 to max_address
			size64 = size64 - 255
			self.eraseFlash(handle, board.memstart + 0x4000, size64)

		# write 32-bit blocks
		# ----------------------------------------------------------------------

		usbBuf = []
		for addr in range(board.memstart, max_address):
			index = addr - board.memstart
			#print hex(addr)
			if addr % self.DLN_BLOCKSIZE == 0:
				if usbBuf != []:
					#print usbBuf
					self.writeFlash(handle, addr - self.DLN_BLOCKSIZE, usbBuf)
				usbBuf = []
			if data[index] != []:
				#print data[addr - memstart]
				usbBuf.append(data[index])
		#print "%d bytes written." % codesize

		return self.ERR_NONE
# ------------------------------------------------------------------------------
	def writeHex(self, output, filename, board):
# ------------------------------------------------------------------------------
		fichier = open(filename, 'r')
		if fichier == "":
			self.add_report(output, "Unable to open %s" % filename)
			return
		fichier.close()

		device = self.getDevice(board)
		if device == self.ERR_DEVICE_NOT_FOUND:
			self.add_report(output, "Pinguino not found")
			self.add_report(output, "Is your device connected and/or in bootloader mode ?")
			return
		else:
			self.add_report(output, "Pinguino found")

		handle = self.initDevice(device)
		if handle == self.ERR_USB_INIT1:
			self.add_report(output, "Upload not possible")
			self.add_report(output, "Try to restart the bootloader mode")
			return

		device_id = self.getDeviceID(handle)
		proc = self.getDeviceName(device_id)
		if proc != board.proc:
			self.add_report(output, "Compiled for %s but device has %s" % (board.proc, proc))
			return
		#memend = self.getDeviceFlash(device_id)
		self.add_report(output, "%s (id=%s)" % (proc, hex(device_id)))
		self.add_report(output, "%d bytes free" % (board.memend - board.memstart))

		#product = handle.getString(device.iProduct, 30)
		#manufacturer = handle.getString(device.iManufacturer, 30)
		self.add_report(output, "Pinguino HID bootloader %s" % self.getVersion(handle))

		if filename == '':
			self.add_report(output, "No program to write")
			self.closeDevice(handle)
			return

		self.add_report(output, "Writing ...")
		status = self.hexWrite(handle, filename, board)
		if status == self.ERR_NONE:
			self.add_report(output, os.path.basename(filename) + " successfully uploaded")
		if status == self.ERR_HEX_RECORD:
			self.add_report(output, "Record error")
			self.closeDevice(handle)
			return
		if status == self.ERR_HEX_CHECKSUM:
			self.add_report(output, "Checksum error")
			self.closeDevice(handle)
			return
		if status == self.ERR_USB_ERASE:
			self.add_report(output, "Erase error")
			self.closeDevice(handle)
			return

		# Disable EIS and Start User App.
		self.add_report(output, "Resetting ...")
		self.reset(handle)
		self.closeDevice(handle)
		return
# ------------------------------------------------------------------------------
